// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "autoconfiguredialog.h"
#include "ui_autoconfiguredialog.h"

#include "farmSettingsNames.h"


AutoConfigureDialog::AutoConfigureDialog(Farm* farm, QWidget *parent) :
    CommonDialog(parent),
    ui(new Ui::AutoConfigureDialog)
{
    ui->setupUi(this);
    setWindowFlags(Qt::FramelessWindowHint | Qt::Dialog);
    update();
    setWindowMask( "", ui->frame->width(), ui->frame->height());

    //////////////////////////////////
    //
    //  Ported section begins here

    this->farm = farm;

    // Creating config collector and parser
    configCollector = new AppPortalConfigCollector(this);
    configParser = new AppPortalConfigParser(this);

    // Application global settings
    m_appSettings = SharedApplicationSettings(new ApplicationSettings(APPLICATION_CONFIG_FILE_NAME));
    m_appSettings->beginGroup(FARM_OVERRIDE_VALUES);

    // Hide browse button
    ui->browseButton->setVisible(false);

    // Connecting signals with slots
    connect(ui->protocolComboBox, SIGNAL(currentIndexChanged(int)), SLOT(currentProtocolChanged(int)));
    connect(ui->browseButton, SIGNAL(clicked()), SLOT(browseFile()));
    connect(ui->manuallyRadioButton, SIGNAL(toggled(bool)), SLOT(manuallyChecked(bool)));
    connect(ui->connectButton, SIGNAL(clicked()), SLOT(accept()));
    connect(ui->cancelButton, SIGNAL(clicked()), SLOT(cancel()));
    connect(configCollector, SIGNAL(networkReply(QNetworkReply::NetworkError, QString, QString)),
            this, SLOT(networkReply(QNetworkReply::NetworkError, QString, QString)));


    this->loadSettings();



}

AutoConfigureDialog::~AutoConfigureDialog()
{
    delete ui;
}

void AutoConfigureDialog::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}


QString AutoConfigureDialog::buildUrl( void )
{
    QString url;

    url.clear();
    url.append( ui->protocolComboBox->currentText());
    url.append( ui->urlEdit->text());

    // Check for the seperator
    if ( !url.endsWith("/")) url.append( "/" );

    url.append( ui->fileEdit->text());

    syslog( LOG_INFO, "Autoconfigure built a URL:'%s'", url.toAscii().data());

    return url;
}

void AutoConfigureDialog::on_connectButton_clicked()
{
    QString url;

    url = buildUrl();

}


////////////////////////////////////////////////////////////////
//
// Ported from old LoadConfiguration Dialog
//
////////////////////////////////////////////////////////////////


void AutoConfigureDialog::currentProtocolChanged(int index)
{
    // If file then show browse button
    QString currentProtocol = ui->protocolComboBox->itemText(index);
    bool browseButtonVisible;

    if (currentProtocol.compare("FILE", Qt::CaseInsensitive) == 0)
    {
        browseButtonVisible = true;
        ui->urlLabel->setText(tr("Dir:"));
        ui->proxyGroupBox->setEnabled(false);
    }
    else
    {
        browseButtonVisible = false;
        ui->urlLabel->setText(tr("URL:"));
        ui->proxyGroupBox->setEnabled(true);
    }
    ui->browseButton->setVisible(browseButtonVisible);
}


void AutoConfigureDialog::browseFile()
{
    // Getting file name from dialog
    QString absName = QFileDialog::getOpenFileName(this, tr("Open File"), "/home", tr("XML files (*.xml)"));
    // Nothing actions if name = 0
    if (absName.isEmpty())
    {
        return;
    }

    // Parse name
    int index = absName.lastIndexOf('/');
    QString dir = absName.mid(0, index);
    QString name = absName.mid(index + 1, absName.length() - index);

    // Add dirname and filename to edits
    ui->urlEdit->setText(dir);
    ui->fileEdit->setText(name);
}

void AutoConfigureDialog::manuallyChecked(bool checked)
{
    //If manually internet settings is checked
    ui->proxyEdit->setEnabled(checked);

    // If manually isn`t checked then clear proxy edit
    if (!checked)
    {
        ui->proxyEdit->setText("");
    }
}

void AutoConfigureDialog::accept()
{
    // Loading farm
    bool result = this->loadFarm();
    // If farm from FILE was parsed successfully

    if(result)
    {
        // Saving farm settings to next settings
        syslog( LOG_INFO, "Autoconfigure Farm: Successfully loaded a settings into memory");
        this->saveSettings();
        return QDialog::accept();
    }
}

void AutoConfigureDialog::cancel()
{
    return QDialog::reject();
}

bool AutoConfigureDialog::loadFarm()
{
    QString currentProtocol = ui->protocolComboBox->currentText();
    ConfigResult result;
    BasicMessageDialog messageBox;

    if (currentProtocol.compare("FILE", Qt::CaseInsensitive) == 0)
    {
        result = configCollector->loadConfigFromFile(ui->urlEdit->text(), ui->fileEdit->text());
    }
    else
    {
        result = configCollector->loadConfigFromNetwork(currentProtocol, ui->urlEdit->text(), ui->fileEdit->text(),
                                                        ui->manuallyRadioButton->isChecked(), ui->proxyEdit->text());
    }
    switch(result)
    {

        case CONFIG_DIR_IS_EMPTY:
            // QMessageBox::critical(this, tr("Configuration error"), tr("Please, select file"));
            messageBox.initialize( tr("Configuration error"), tr("Please, select file" ));
            messageBox.exec();
            break;
        case CONFIG_URL_IS_EMPTY:
            // QMessageBox::critical(this, tr("URL error"), tr("Enter correct URL"));
            messageBox.initialize( tr("URL error"), tr("Enter correct URL") );
            messageBox.exec();
            break;
        case CONFIG_FILE_NAME_IS_EMPTY:
            // QMessageBox::critical(this, tr("File error"), tr("Enter correct file name"));
            messageBox.initialize( tr("File error"), tr("Enter correct file name"));
            messageBox.exec();
            break;
        case CONFIG_FILE_DOES_NOT_EXIST:
            // QMessageBox::critical(this, tr("File read error"), tr("File doesn`t exist!"));
            messageBox.initialize(tr("File read error"), tr("File doesn`t exist!") );
            messageBox.exec();
            break;
        case CONFIG_FILE_COULD_NOT_BE_READ:
            // QMessageBox::critical(this, tr("File read error"), tr("File couldn`t be read!"));
            messageBox.initialize( tr("File read error"), tr("File couldn`t be read!"));
            messageBox.exec();
            break;
        case CONFIG_WRONG_PROXY_PORT:
            // QMessageBox::critical(this, tr("Connection error"), tr("Enter correct value of proxy port"));
            messageBox.initialize( tr("Connection error"), tr("Enter correct value of proxy port"));
            messageBox.exec();
            break;
        case CONFIG_OK:
            if (currentProtocol.compare("FILE", Qt::CaseInsensitive) == 0)
            {
                return parseConfigFromFile(ui->urlEdit->text(), ui->fileEdit->text());
            }
            break;



    }
    return false;
}

bool AutoConfigureDialog::parseConfigFromFile(QString dir, QString fileName)
{
    QString protocol = "file";
    SharedBuffer sharedBuffer = configCollector->getConfigBuffer(protocol, dir, fileName);
    if (sharedBuffer)
    {
        bool result = configParser->parseFromBuffer(farm, *sharedBuffer);
        if (result)
        {
            QMessageBox::information(this, this->windowTitle(), tr("Configuration [") + farm->getName() +
                         tr("] loaded successfully from [") + ui->urlEdit->text() + "/" + ui->fileEdit->text() +
                         tr("]."));
            return true;
        }
        else
        {
            QMessageBox::information(this, this->windowTitle(), tr("The following error occured while "
                    "loading the farm definition from [") + ui->urlEdit->text() + "/" + ui->fileEdit->text() +
                    tr("]: Invalid configuration file.\nPlease specify a valid file for farm definition."));
        }
    }
    else
    {
        QMessageBox::critical(this, tr("Config error"), tr("There is no config: ") +
                              ui->urlEdit->text() + "/" + ui->fileEdit->text());
    }
    return false;
}

void AutoConfigureDialog::networkReply(QNetworkReply::NetworkError error, QString errorString, QString urlString)
{
    if (error == QNetworkReply::NoError)
    {
        // Parsing data
        SharedBuffer sharedBuffer = configCollector->getConfigBuffer(urlString);
        if (sharedBuffer)
        {
            if (!configParser->parseFromBuffer(farm, *sharedBuffer))
            {
                QMessageBox::critical(this, tr("Parsing error"), tr("Error in parsing data"));
            }
            else
            {
                QMessageBox::information(this, this->windowTitle(), tr("Configuration [") + farm->getName() +
                             tr("] loaded successfully from [") + urlString + tr("]."));
                this->saveSettings();
                return QDialog::accept();
            }
        }
        else
        {
            QMessageBox::critical(this, tr("Config error"), tr("There is no config: ") + urlString);
        }
    }
    else
    {
        QMessageBox::critical(this, tr("Connection error"), tr("Error occured while loading\n[") + urlString +
                                    tr("]\nError code: ") + errorString);
    }
}

void AutoConfigureDialog::saveSettings()
{
    farm->setSourceProtocol(ui->protocolComboBox->currentText().toLower());
    farm->setSourceServer(ui->urlEdit->text());
    farm->setSourceFile(ui->fileEdit->text());

    farm->setSourceProxy(ui->proxyEdit->text());
}


void AutoConfigureDialog::loadSettings()
{
    int protocolIndex = ui->protocolComboBox->findText(farm->getSourceProtocol(), Qt::MatchFixedString);
    if (protocolIndex >= 0)
    {
        ui->protocolComboBox->setCurrentIndex(protocolIndex);
    }

    ui->urlEdit->setText(farm->getSourceServer());
    ui->fileEdit->setText(farm->getSourceFile());

    QString proxy = farm->getSourceProxy();
    if (proxy.isEmpty())
    {
        ui->defaultRadioButton->setChecked(true);
    }
    else
    {
        ui->manuallyRadioButton->setChecked(true);
        ui->proxyEdit->setText(proxy);
    }

    QString tmpStr;
    if (m_appSettings->getValue(SOURCE_PROTOCOL, &tmpStr) && protocolIndex >= 0)
    {
        ui->protocolComboBox->setEnabled(false);
    }
    if (m_appSettings->getValue(SOURCE_SERVER, &tmpStr))
    {
        ui->urlEdit->setEnabled(false);
        ui->browseButton->setEnabled(false);
    }
    ui->fileEdit->setEnabled(!m_appSettings->getValue(SOURCE_FILE, &tmpStr));
    if (m_appSettings->getValue(SOURCE_PROXY, &tmpStr))
    {
        ui->proxyEdit->setEnabled(false);
        if (proxy.isEmpty())
        {
            ui->manuallyRadioButton->setEnabled(false);
        }
        else
        {
            ui->defaultRadioButton->setEnabled(false);
        }
    }
}


void AutoConfigureDialog::on_backButton_clicked()
{
    // Return to the Add Farm dialog.
    this->setResult( 2 );
    hide();
}
void AutoConfigureDialog::on_cancelButton_clicked()
{
    this->reject();
}
