// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "farmConnection.h"
#include "ChangeLocationDialog/changeLocationDialog.h"
#include "CredentialsDialog/credentialsDialog.h"
#include "ChangePasswordDialog/changePasswordDialog.h"
#include "common/errors.h"
#include "restartProgressDialog.h"
#include "sublib/code.h"

#include <syslog.h>
#include <QtDebug>

QString getPntscErrorString(int errcode)
{
    switch (errcode){
    case NO_ERROR:
        return QString("Operation succeeded");
    case INTERNAL_ERROR:
        return QString("Internal program error");
    case INPUT_PARAMETERS_ERROR:
        return QString("Incorrect input parameters");
    case SYSTEM_ERROR:
        return QString("System error");
    case AUTHENTICATION_ERROR:
        return QString("Authentication error");
    case BROKER_ERROR:
    default: // usually all other codes are errors returned from broker
            // besides this, if I were the user I would like to get
            // a 'broker error' rather than an 'unknown' one
        return QString("Broker error");
    }
}

// ---

FarmConnection::FarmConnection(QObject * parent/* = 0*/)
        : QObject(parent),
        pntsc(0),
        p_farm(0),
        isAutoLaunched(false),
        p_connected(false),
        p_requestCreds(false)
{

}

/*virtual*/ FarmConnection::~FarmConnection()
{
    // Stop the process
    /*if (pntsc != 0)
    {
        pntsc->kill();
        delete pntsc;
        pntsc = 0;
    }*/
}

Farm * FarmConnection::farm() const
{
    return p_farm;
}

void FarmConnection::setFarm(Farm *farm)
{
    p_farm = farm;

    //qDebug() << p_farm->getDefaultLocation();

    if (p_farm == 0)
        return;

    // Get last credentials
    p_username = p_farm->getUsername();
    p_password = p_farm->getPassword();
    p_domain = p_farm->getCurrentDomain();
}

bool FarmConnection::isConnected() const
{
    return p_connected;
}

void FarmConnection::setConnected(bool connected)
{
    p_connected = connected;
}

QString FarmConnection::lastError() const
{
    return p_lastError;
}

void FarmConnection::reject()
{
    if (pntsc == 0)
        return;

    // Stoping the process
    pntsc->terminate();
    pntsc->waitForFinished();

    //delete pntsc;
    //pntsc = 0;
}

void FarmConnection::readAvailableDataOnRefreshing()
{
    QString currentData = QString(pntsc->readLine()).trimmed();

    // Empty data was received
    if (currentData.isEmpty())
        return;

    else if (currentData.startsWith("-"))
    {
        QString statusMessage = currentData.remove("-").trimmed();
        return emit stateChanged(statusMessage);
    }
}

void FarmConnection::readAvailableDataOnStart()
{
    // Reading available data
    QString currentData = QString(pntsc->readAllStandardOutput()/*readLine()*/);
    QStringList list = currentData.split("\n", QString::SkipEmptyParts);

    // Empty data was received
    if (currentData.trimmed().isEmpty() || list.isEmpty())
        return;

    // Current data is status message
    if (list.first().startsWith("POWERING_UP"))
    {
        const int timeoutOffset = 1;
        const int attemptOffset = 2;
        const int msgOffset = 3;
        const QString defaultMsg(tr("The computer operating system is loading..."));
        const QString plsWaitMsg(tr(" Please wait."));

        QString message;
        int timeout = list.at(timeoutOffset).toInt();
        int curAttemptNum = list.at(attemptOffset).toInt();
        message = list.at(msgOffset);
        if (message.isEmpty())
        {
            message = defaultMsg;
        }
        message += plsWaitMsg;

        RestartProgressDialog *restartProgressDialog = RestartProgressDialog::getInstance(&pntsc, message, timeout, curAttemptNum);
        restartProgressDialog->show();
    }
    else if (!list.last().startsWith("-") && list.last() != "")
    {
        QString statusMessage = currentData.remove("-").trimmed();
        return emit stateChanged(statusMessage);
    }
}

QProcessEnvironment FarmConnection::getEnvironmentWithUtf8()
{
    const QString defaultLang = "en_US.UTF-8";

    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
    QString lang = env.value("LANG"); //linux only, not platform independent

    int dotIndex = lang.indexOf('.');
    if (-1 == dotIndex)
    {
        lang = defaultLang;
    }
    else
    {
        lang.truncate(dotIndex + 1);
        lang.append("UTF8");
    }

    env.insert("LANG", lang);

    return env;
}

void FarmConnection::changePasswordRequest(const QString & newPassword)
{
    // Clear last errors
    p_lastError.clear();

    if (!p_farm->isPasswordManagementAllowed())
    {
        p_lastError = tr("Password Management server is not allowed.");
        return;
    }

    // Sending query
    QStringList request = this->createChangePasswordRequest(newPassword);
    if (request.isEmpty())
    {
        p_lastError = tr("Password change request was not created.");
        return;// false;
    }

    // Create new process
    pntsc = new QProcess(this->parent());

    // Connecting process with slots
    connect(pntsc, SIGNAL(finished(int, QProcess::ExitStatus)),
                SLOT(changingPasswordFinished(int, QProcess::ExitStatus)));
    connect(pntsc, SIGNAL(readyReadStandardOutput()), SLOT(readAvailableDataOnRefreshing()));

    // Startint the process
    //qDebug() << "Server response:" << request.join(" ");
    QProcessEnvironment env = getEnvironmentWithUtf8();
    pntsc->setProcessEnvironment(env);
    pntsc->start("pntsc", request);

    // List of current farm application is cleared
    if (!pntsc->waitForStarted())
    {
        p_lastError = tr("Unable to start PNTSC");
        emit error();
        pntsc->kill();
    }
    else
    {
        emit started();
        while (pntsc->state() == QProcess::Running)
                QApplication::processEvents();
    }

    // Delete process
    delete pntsc;
    pntsc = 0;
}

bool FarmConnection::refreshApplicationsList()
{
    // Clear last errors
    p_lastError.clear();

    // Sending query
    QStringList request = this->processRequest();


    if (p_lastError.startsWith(getPntscErrorString(NO_ERROR)))
    {
        return true;
    }

    if (request.isEmpty())
    {
        p_lastError = tr("Unable to process application list request");
        syslog( LOG_WARNING, "Unable to process application list request");
        return false;
    }

    p_applications.clear();

    // Create new process
    pntsc = new QProcess(this->parent());

    // Connecting process with slots
    connect(pntsc, SIGNAL(finished(int, QProcess::ExitStatus)),
                SLOT(refreshingFinished(int, QProcess::ExitStatus)));
    connect(pntsc, SIGNAL(readyReadStandardOutput()), SLOT(readAvailableDataOnRefreshing()));

    // Add application option to request
    request << "-L";

    // Startint the process
    //qDebug() << "List request:" << request.join(" ");
    QProcessEnvironment env = getEnvironmentWithUtf8();
    pntsc->setProcessEnvironment(env);
    pntsc->start("pntsc", request);

    bool result;
    if (!pntsc->waitForStarted())
    {
        p_lastError = tr("Unable to start PNTSC");
        pntsc->kill();
        emit error();
        result = false;
    }
    else
    {
        emit started();
        // Processing GUI events
        while (pntsc->state() == QProcess::Running)
            QApplication::processEvents();
        result = true;
    }

    // Delete process
    delete pntsc;
    pntsc = 0;

    if (p_farm->autoConnected() && !this->isAutoLaunched)
    {
        this->isAutoLaunched = true;
        this->runNextApplication();
    }

    return result;
}

void FarmConnection::executeSelectedApplication(const QString & applicationName)
{
    // Clear last errors
    p_lastError.clear();

    // This mode is enabled when farm is connected
    if (!this->isConnected())
        return;

    // Create pit file with settings
    FileSharedPointer pitFile = p_farm->createPitFile();
    //qDebug() << "Pit-file name:" << pitFile->fileName();

    if (pitFile == 0)// || pitFile->open(QIODevice::ReadOnly))
    {
        p_lastError = tr("Error creating pit-file for this connection");
        emit error();
        return;
    }

    // Sending query
    QStringList request = this->processRequest();
    if (request.isEmpty())
    {
        p_lastError = tr("Unable to process application start request");
        return;// false;
    }

    // Create new process
    pntsc = new QProcess(this->parent());

    // Connecting process with slots
    connect(pntsc, SIGNAL(finished(int, QProcess::ExitStatus)),
                SLOT(runningApplicationFinished(int, QProcess::ExitStatus)));
    connect(pntsc, SIGNAL(readyReadStandardOutput()), SLOT(readAvailableDataOnStart()));

    // Add running option to request
    request << "-a" << applicationName
            << "-f" << pitFile->fileName();

    // Startint the process
    QProcessEnvironment env = getEnvironmentWithUtf8();
    pntsc->setProcessEnvironment(env);
    pntsc->start("pntsc", request);
    //qDebug() << "Run app request:" << request;

    // List of current farm application is cleared
    if (!pntsc->waitForStarted())
    {
        p_lastError = tr("Unable to start PNTSC");
        emit error();
        pntsc->kill();
    }
    else
    {
        emit started();
        while (pntsc != NULL && pntsc->state() == QProcess::Running)
                QApplication::processEvents();
    }
}

bool FarmConnection::connectTo()
{
    // Disconnecting with old farm
    this->setConnected(false);

    // Clear last errors
    p_lastError.clear();

    // Farm doesn`t exist
    if (p_farm == 0)
    {
        p_lastError = tr("Connection farm was not selected");
        return false;
    }

    // Break if process is running

    // It was commented to fix issue "Cannot connect to the farm after unsuccessful launching session".
    // Here refreshApplicationsList() is permitted to get again the list with applications.
    /*
    if (pntsc != 0)
    {
        p_lastError = tr("Connection is busy");
        return false;
    }
    */

    // Auto launch list initialization
    p_autoLaunchList = p_farm->autolaunch().names();
    //qDebug() << "connecTo().autolaunchlist:" << p_farm->applications().names();

    // Sending request to update applications list
    return this->refreshApplicationsList();
}

void FarmConnection::disconnectTo()
{
    // TODO:: all method
    /*
    // Set off connection flag
    this->setState(Empty);

    // Clear last errors
    p_lastError.clear();

    // List of current farm application is cleared
    p_appNames.clear();
*/
}

bool FarmConnection::changeCurrentLocation()
{
    // This mode is enabled when farm is connected
    if (!this->isConnected())
        return false;

    // Show change location dialog
    ChangeLocationDialog *changeLocationDialog = new ChangeLocationDialog(p_farm,
                                                                          (QWidget *) this->parent());

    // Changing was canceled
    if (changeLocationDialog->exec() == QDialog::Rejected)
                return false;

    // Sending request to update applications list
    return this->refreshApplicationsList();
}

bool FarmConnection::logonAsADifferentUser()
{
    // This mode is enabled when farm is connected
    if (!this->isConnected())
        return false;

    // Show credentials dialog
    CredentialsDialog *credentialsDialog = new CredentialsDialog(&p_username,
                                                                 &p_domain,
                                                                 &p_password,
                                                                 p_farm,
                                                                 // (QWidget *) this->parent());
                                                                 NULL );


    QWidget * mw = (QWidget *) this->parent();
    credentialsDialog->move( QPoint( (mw->x() + mw->width()/2) - credentialsDialog->width()/2,
                                   ( mw->y() + mw->height()/2) - credentialsDialog->height()/2 ));

    // Changing was canceled
    if (credentialsDialog->exec() == QDialog::Rejected) return false;

    // Sending request to update applications list
    return this->refreshApplicationsList();
}

bool FarmConnection::changePassword(FarmSettings *settings)
{
    bool passManAllowed = false;
    for(int i = 0; i < settings->getFarmNamesList().count(); ++i)
    {
        if (settings->getFarm(i)->isPasswordManagementAllowed() && !settings->getFarm(i)->getPasswordManagementServer().isEmpty())
        {
            passManAllowed = true;
            break;
        }
    }

    if (!passManAllowed)
    {
        QMessageBox::information(0, tr("Change Password"), tr("There are no farms with Password Managment allowed."));
        return false;
    }

    // Show credentials dialog
    QString newPassword;

    ChangePasswordDialog *passwordDialog = new ChangePasswordDialog(settings, &p_username,
                                                                 &p_domain,
                                                                 &p_password,
                                                                 &newPassword,
                                                                 &p_farm,
                                                                 //(QWidget *) this->parent()
                                                                 NULL
                                                                 );

    QWidget * mw = (QWidget *) this->parent();
    passwordDialog->move( QPoint(( mw->x() + mw->width()/2) - passwordDialog->width()/2,
                                 ( mw->y() + mw->height()/2) - passwordDialog->height()/2 ));

    // Changing was canceled
    if (passwordDialog->exec() == QDialog::Rejected)
        return false;

    // Sending request to change password
    this->changePasswordRequest(newPassword);
    return true;
}

void FarmConnection::refreshingFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    //qDebug() << "Exit code:" << exitCode;

    // Process was crashed
    if (exitStatus == QProcess::CrashExit)
    {
        //p_lastError = tr("Operation was canceled...");
        //emit error();
        return emit finished(Reject);
    }

    // Wasn`t successfully
    if (exitCode != 0)
    {
        QByteArray data = pntsc->readAllStandardError();

        // parse received error string
        QStringList list = QString(data).split("],");
        QString code = list.at(0);
        QString module = list.at(1);
        QString inf = list.at(2);
        code.replace("[ERR","");
        module.replace("[","");

        bool ok;
        int errcode = code.toInt(&ok, 16);

        if (errcode != UNKNOWN_ERROR) //this corresponds that the application's list is empty
        {
        QString errcodestr = getPntscErrorString(errcode);
        p_lastError = "<b>[Error " + code + "]:</b> " + errcodestr +
            "<p><b>[Module]:</b> " + module +
            "<p><b>[Description]:</b> " + inf;

        return emit finished(Error);
        }
    }


    // Reading output
    QByteArray outputData = pntsc->readAllStandardOutput();
    QStringList list = QString(outputData).split("\n", QString::SkipEmptyParts);

    syslog( LOG_INFO, "Application string: %s", outputData.data());
    syslog( LOG_INFO, "Loaded %i applications/desktops", list.count());

    if (list.isEmpty())
    {
        const QString noAvailableApps("Name=" + tr("No available applications") + ",FolderName=" + tr("My Applications") + ",Enabled=false");
        list.push_back(noAvailableApps);
    }

    p_applications.setApplications(list);

    // TODO: find the place of this item
    p_farm->setApplications(p_applications);

    // Connection successfully completed
    this->setConnected(true);
    return emit finished(Success);
}

void FarmConnection::runningApplicationFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    //qDebug() << "RunningApplicationFinished().Exit code:" << exitCode;

    // Process was crashed
    if (exitStatus == QProcess::CrashExit)
    {
        //p_lastError = tr("Operation was canceled...");
        return emit finished(Reject);
    }

    // Wasn`t successfully
    if (exitCode != 0)
    {
        int errcode = UNKNOWN_ERROR;
        QString code("unknown error");
        QString module("pntsc");
        QString inf("PNTSC exited unexpectedly - check your connection information");
        if (pntsc != 0)
        {
            QByteArray data = pntsc->readAllStandardError();
            if (!data.isNull() && !data.isEmpty())
            {
                // parse received error string
                QStringList list = QString(data).split("],");
                code = list[0];
                module = list[1];
                inf = list[2];
                code.replace("[ERR","");
                module.replace("[","");

                bool ok;
                errcode = code.toInt(&ok, 16);
            }
        }
        QString errcodestr = getPntscErrorString(errcode);
        p_lastError = "[Error " + code + "]: " + errcodestr +
            "\n[Module]: " + module +
            "\n[Description]: " + inf;

        return emit finished(Error);
    }

    // Delete process
    delete pntsc;
    pntsc = 0;

    return emit finished(Success);
}

void FarmConnection::changingPasswordFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    //qDebug() << "ChangingPasswordFinished().Exit code:" << exitCode;

    // Process was crashed
    if (exitStatus == QProcess::CrashExit)
    {
        //p_lastError = tr("Operation was canceled...");
        return emit changePasswordFinished(Reject, this);
    }

    // Wasn`t successfully
    if (exitCode != 0)
    {
        QByteArray data = pntsc->readAllStandardError();

        // parse received error string
        QStringList list = QString(data).split("],");
        QString code = list.at(0);
        QString module = list.at(1);
        QString inf = list.at(2);
        code.replace("[ERR","");
        module.replace("[","");

        bool ok;
        int errcode = code.toInt(&ok, 16);
        QString errcodestr = getPntscErrorString(errcode);
        p_lastError = "[Error " + code + "]: " + errcodestr +
            "\n[Module]: " + module +
            "\n[Description]: " + inf;

        return emit changePasswordFinished(Error, this);
    }

    return emit changePasswordFinished(Success, this);
}

QStringList FarmConnection::processRequest()
{
    QStringList request;

    LocationSharedPointer currentLocation = p_farm->getCurrentLocation();
    if (currentLocation == 0)
    {
        syslog( LOG_ERR, "Error: No current location has been set");
        return request;
    }

    // Treating location settings
    QString servers = currentLocation->getServers().join(",");
    QString port = currentLocation->getPort();
    QString protocol = currentLocation->getProtocol();

    // Not correctly  values
    if (servers.isEmpty() || protocol.isEmpty())
    {
        syslog( LOG_ERR, "Error: Connections settings are not valid.");
        return request;
    }

    QString brokers = protocol + "://" + servers + ":" + port;

    QString sslGateway = currentLocation->getSslGateway();
    bool isSsl = currentLocation->isSsl();
    bool isNat = currentLocation->isNat();

    // Adding options into request
    request << brokers;
    if (isNat)
        request << "-t";
    if (isSsl)
        request << "-e" << sslGateway;

    QString proxy = currentLocation->getProxyServer();
    if (proxy != "")
        request << "-P" << proxy;

    // Adding credentials into request
/*
    if (!p_username.isEmpty() && !p_password.isEmpty() && !p_domain.isEmpty())
        request << "-u" << p_username << "-p" << p_password
                << "-d" << p_domain;
*/
    if (p_username.isEmpty() || p_password.isEmpty() || p_domain.isEmpty() || p_requestCreds)
    {
        p_requestCreds = false;
        CredentialsDialog *credentialsDialog = new CredentialsDialog(&p_username,
                                                                 &p_domain,
                                                                 &p_password,
                                                                 p_farm,
                                                              //   (QWidget *) this->parent());
                                                                 NULL );

        // Changing was canceled
        if (credentialsDialog->exec() == QDialog::Rejected)
        {
            request.clear();
            p_lastError = getPntscErrorString(NO_ERROR);;
            return request;
        }

    }

    request << "-u" << ( p_username ) << "-p" << ( p_password ) << "-d" << ( p_domain );

    return request;
}

void FarmConnection::setRequestCredentials(bool requestCreds)
{
    this->p_requestCreds = requestCreds;
}

QStringList FarmConnection::createChangePasswordRequest(const QString & newPassword)
{
    QStringList request;

    // Password management is dissalowed
    if (!p_farm->isPasswordManagementAllowed())
        return request;

    // Treating location settings
    QString servers = p_farm->getPasswordManagementServer();
    int port = p_farm->getPasswordManagementPort();
    // TODO: is it true?
    QString protocol = "https";

    // Not correctly  values
    if (servers.isEmpty() || protocol.isEmpty())
        return request;

    QString brokers = protocol + "://" + servers + ":" + QString::number(port);
    request << brokers;

    // Adding credentials into request
    if (!p_username.isEmpty() && !p_password.isEmpty() && !p_domain.isEmpty())
        request << "-u" << p_username << "-p" << p_password
                << "-d" << p_domain << "-c" << newPassword;

    return request;
}

void FarmConnection::runNextApplication()
{
    QString currentApplicationName;

    if (p_autoLaunchList.isEmpty())
    {
        if (p_farm->autoConnectIfOneApplication() &&
            (p_farm->applications().size() == 1))
        {
            // Try to run first application from application list
            currentApplicationName = p_farm->applications().names().takeFirst();
        }
        else
        {
            // There are no available applications to run
            return;
        }
    }
    else
    {
        // Try to run first application from auto-launch list
        currentApplicationName = p_autoLaunchList.takeFirst();
    }

    emit startApplication(currentApplicationName);
}

ApplicationList FarmConnection::applications() const
{
    return p_applications;
}
