// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "loadConfigurationDialog.h"

#include <QtDebug>
#include <QFileDialog>
#include "common/common.h"
#include "farmSettingsNames.h"

LoadConfigurationDialog::LoadConfigurationDialog(Farm* farm, QWidget* parent /*= 0*/)
    : QDialog(parent)
{
    setupUi(this);

    this->farm = farm;

    // Creating config collector and parser
    configCollector = new AppPortalConfigCollector(this);
    configParser = new AppPortalConfigParser(this);

    // Application global settings
    m_appSettings = SharedApplicationSettings(new ApplicationSettings(APPLICATION_CONFIG_FILE_NAME));
    m_appSettings->beginGroup(FARM_OVERRIDE_VALUES);

    // Hide browse button
    browseButton->setVisible(false);

    // Connecting signals with slots
    connect(protocolComboBox, SIGNAL(currentIndexChanged(int)), SLOT(currentProtocolChanged(int)));
    connect(browseButton, SIGNAL(clicked()), SLOT(browseFile()));
    connect(manuallyRadioButton, SIGNAL(toggled(bool)), SLOT(manuallyChecked(bool)));
    connect(okButton, SIGNAL(clicked()), SLOT(accept()));
    connect(cancelButton, SIGNAL(clicked()), SLOT(cancel()));
    connect(configCollector, SIGNAL(networkReply(QNetworkReply::NetworkError, QString, QString)),
            this, SLOT(networkReply(QNetworkReply::NetworkError, QString, QString)));

    this->loadSettings();
}

LoadConfigurationDialog::~LoadConfigurationDialog()
{
}

void LoadConfigurationDialog::currentProtocolChanged(int index)
{
    // If file then show browse button
    QString currentProtocol = protocolComboBox->itemText(index);
    bool browseButtonVisible;

    if (currentProtocol.compare("FILE", Qt::CaseInsensitive) == 0)
    {
        browseButtonVisible = true;
        urlLabel->setText(tr("Dir:"));
        proxyGroupBox->setEnabled(false);
    }
    else
    {
        browseButtonVisible = false;
        urlLabel->setText(tr("URL:"));
        proxyGroupBox->setEnabled(true);
    }
    browseButton->setVisible(browseButtonVisible);
}

void LoadConfigurationDialog::browseFile()
{
    // Getting file name from dialog
    QString absName = QFileDialog::getOpenFileName(this, tr("Open File"), "/home", tr("XML files (*.xml)"));
    // Nothing actions if name = 0
    if (absName.isEmpty())
    {
        return;
    }

    // Parse name
    int index = absName.lastIndexOf('/');
    QString dir = absName.mid(0, index);
    QString name = absName.mid(index + 1, absName.length() - index);

    // Add dirname and filename to edits
    urlEdit->setText(dir);
    fileEdit->setText(name);
}

void LoadConfigurationDialog::manuallyChecked(bool checked)
{
    //If manually internet settings is checked
    proxyEdit->setEnabled(checked);

    // If manually isn`t checked then clear proxy edit
    if (!checked)
    {
        proxyEdit->setText("");
    }
}

void LoadConfigurationDialog::accept()
{
    // Loading farm
    bool result = this->loadFarm();
    // If farm from FILE was parsed successfully
    if(result)
    {
        // Saving farm settings to next settings
        this->saveSettings();
        return QDialog::accept();
    }
}

void LoadConfigurationDialog::cancel()
{
    return QDialog::reject();
}

bool LoadConfigurationDialog::loadFarm()
{
    QString currentProtocol = protocolComboBox->currentText();
    ConfigResult result;
    if (currentProtocol.compare("FILE", Qt::CaseInsensitive) == 0)
    {
        result = configCollector->loadConfigFromFile(urlEdit->text(), fileEdit->text());
    }
    else
    {
        result = configCollector->loadConfigFromNetwork(currentProtocol, urlEdit->text(), fileEdit->text(),
                                                        manuallyRadioButton->isChecked(), proxyEdit->text());
    }
    switch(result)
    {
        case CONFIG_DIR_IS_EMPTY:
            QMessageBox::critical(this, tr("Configuration error"), tr("Please, select file"));
            break;
        case CONFIG_URL_IS_EMPTY:
            QMessageBox::critical(this, tr("URL error"), tr("Enter correct URL"));
            break;
        case CONFIG_FILE_NAME_IS_EMPTY:
            QMessageBox::critical(this, tr("File error"), tr("Enter correct file name"));
            break;
        case CONFIG_FILE_DOES_NOT_EXIST:
            QMessageBox::critical(this, tr("File read error"), tr("File doesn`t exist!"));
            break;
        case CONFIG_FILE_COULD_NOT_BE_READ:
            QMessageBox::critical(this, tr("File read error"), tr("File couldn`t be read!"));
            break;
        case CONFIG_WRONG_PROXY_PORT:
            QMessageBox::critical(this, tr("Connection error"), tr("Enter correct value of proxy port"));
            break;
        case CONFIG_OK:
            if (currentProtocol.compare("FILE", Qt::CaseInsensitive) == 0)
            {
                return parseConfigFromFile(urlEdit->text(), fileEdit->text());
            }
            break;
    }
    return false;
}

bool LoadConfigurationDialog::parseConfigFromFile(QString dir, QString fileName)
{
    QString protocol = "file";
    SharedBuffer sharedBuffer = configCollector->getConfigBuffer(protocol, dir, fileName);
    if (sharedBuffer)
    {
        bool result = configParser->parseFromBuffer(farm, *sharedBuffer);
        if (result)
        {
            QMessageBox::information(this, this->windowTitle(), tr("Configuration [") + farm->getName() +
                         tr("] loaded successfully from [") + urlEdit->text() + "/" + fileEdit->text() +
                         tr("]."));
            return true;
        }
        else
        {
            QMessageBox::information(this, this->windowTitle(), tr("The following error occured while "
                    "loading the farm definition from [") + urlEdit->text() + "/" + fileEdit->text() +
                    tr("]: Invalid configuration file.\nPlease specify a valid file for farm definition."));
        }
    }
    else
    {
        QMessageBox::critical(this, tr("Config error"), tr("There is no config: ") +
                              urlEdit->text() + "/" + fileEdit->text());
    }
    return false;
}

void LoadConfigurationDialog::networkReply(QNetworkReply::NetworkError error, QString errorString, QString urlString)
{
    if (error == QNetworkReply::NoError)
    {
        // Parsing data
        SharedBuffer sharedBuffer = configCollector->getConfigBuffer(urlString);
        if (sharedBuffer)
        {
            if (!configParser->parseFromBuffer(farm, *sharedBuffer))
            {
                QMessageBox::critical(this, tr("Parsing error"), tr("Error in parsing data"));
            }
            else
            {
                QMessageBox::information(this, this->windowTitle(), tr("Configuration [") + farm->getName() +
                             tr("] loaded successfully from [") + urlString + tr("]."));
                this->saveSettings();
                return QDialog::accept();
            }
        }
        else
        {
            QMessageBox::critical(this, tr("Config error"), tr("There is no config: ") + urlString);
        }
    }
    else
    {
        QMessageBox::critical(this, tr("Connection error"), tr("Error occured while loading\n[") + urlString +
                                    tr("]\nError code: ") + errorString);
    }
}

void LoadConfigurationDialog::saveSettings()
{
    farm->setSourceProtocol(protocolComboBox->currentText().toLower());
    farm->setSourceServer(urlEdit->text());
    farm->setSourceFile(fileEdit->text());

    farm->setSourceProxy(proxyEdit->text());
}

void LoadConfigurationDialog::loadSettings()
{
    int protocolIndex = protocolComboBox->findText(farm->getSourceProtocol(), Qt::MatchFixedString);
    if (protocolIndex >= 0)
    {
        protocolComboBox->setCurrentIndex(protocolIndex);
    }

    urlEdit->setText(farm->getSourceServer());
    fileEdit->setText(farm->getSourceFile());

    QString proxy = farm->getSourceProxy();
    if (proxy.isEmpty())
    {
        defaultRadioButton->setChecked(true);
    }
    else
    {
        manuallyRadioButton->setChecked(true);
        proxyEdit->setText(proxy);
    }

    QString tmpStr;
    if (m_appSettings->getValue(SOURCE_PROTOCOL, &tmpStr) && protocolIndex >= 0)
    {
        protocolComboBox->setEnabled(false);
    }
    if (m_appSettings->getValue(SOURCE_SERVER, &tmpStr))
    {
        urlEdit->setEnabled(false);
        browseButton->setEnabled(false);
    }
    fileEdit->setEnabled(!m_appSettings->getValue(SOURCE_FILE, &tmpStr));
    if (m_appSettings->getValue(SOURCE_PROXY, &tmpStr))
    {
        proxyEdit->setEnabled(false);
        if (proxy.isEmpty())
        {
            manuallyRadioButton->setEnabled(false);
        }
        else
        {
            defaultRadioButton->setEnabled(false);
        }
    }
}
