// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <stdexcept>
#include "farmSettingsNames.h"
#include "displayPage.h"


DisplayPage::DisplayPage(Farm* farm, ManageConnectionDialog* parent)
    : ManageConnectionPage(farm, parent)
{
    setupUi(this);

    m_desktopSizeList << "640x480"  << "800x480"  << "800x600"
                      << "1024x480" << "1024x600" << "1024x768"
                      << "1280x768" << "1280x800" << "Full Screen";

    m_pixmapFileNames << ":/img/ColorSample256.png"
                      << ":/img/ColorSample15bit.png"
                      << ":/img/ColorSample16bit.png"
                      << ":/img/ColorSample24True.png";

    m_colorDepthList << 8 << 15 << 16 << 24;

    bool connected = true;
    connected &= connect(donePushButton, SIGNAL(clicked()), parent, SLOT(accept()));
    connected &= connect(desktopSizeSlider, SIGNAL(valueChanged(int)), this, SLOT(displaySizeChanged(int)));
    connected &= connect(colorDepthComboBox, SIGNAL(currentIndexChanged(int)), SLOT(colorDepthChanged(int)));
    connected &= connect(displayConnectionBarCheckBox, SIGNAL(toggled(bool)), SLOT(pinConnectionBarEnabled(bool)));
    if (!connected)
    {
        throw std::runtime_error("QT's signal and slot connection error in DisplayPage constructor");
    }

    loadFarmSettings();
}

void DisplayPage::loadFarmSettings()
{
    // Farm doesn`t exist
    if (m_pCurrentFarm == NULL)
    {
        return;
    }

    int width = m_pCurrentFarm->getDesktopWidth();
    int height = m_pCurrentFarm->getDesktopHeight();
    if (height == 0 || width == 0)
    {
        desktopSizeSlider->setValue(desktopSizeSlider->maximum());
    }
    else
    {
        QString sizeString = QString::number(width) + "x" + QString::number(height);
        int sizeIndex = m_desktopSizeList.indexOf(sizeString);
        desktopSizeLabel->setText(sizeString);
        desktopSizeSlider->setValue(sizeIndex);
    }

    int colorDepthIndex = m_colorDepthList.indexOf(m_pCurrentFarm->getColorDepth());
    colorDepthIndex = (colorDepthIndex >= 0) ? colorDepthIndex : 0;
    colorDepthComboBox->setCurrentIndex(colorDepthIndex);
    colorDepthChanged(colorDepthIndex);

    seamlessModeCheckBox->setChecked(m_pCurrentFarm->isSeamlessMode());
    displayConnectionBarCheckBox->setChecked(m_pCurrentFarm->isConnectionBar());
    // pinConnectionBarCheckBox->setChecked(m_pCurrentFarm->isPinConnectionBar());
    multimonCheckBox->setChecked(m_pCurrentFarm->isSpanMonitors());

    int tmpInt;
    bool tmpBool;
    if (m_appSettings->getValue(DESKTOP_WIDTH, &tmpInt) || m_appSettings->getValue(DESKTOP_HEIGHT, &tmpInt))
    {
        desktopSizeSlider->setEnabled(false);
        desktopSizeLabel->setEnabled(false);
        lessLabel->setEnabled(false);
        moreLabel->setEnabled(false);
    }
    colorDepthComboBox->setEnabled(!m_appSettings->getValue(COLOR_DEPTH, &tmpInt));
    seamlessModeCheckBox->setEnabled(!m_appSettings->getValue(SEAMLESS_MODE, &tmpBool));
    displayConnectionBarCheckBox->setEnabled(!m_appSettings->getValue(DISPLAY_CONNECTION_BAR, &tmpBool));
    // pinConnectionBarCheckBox->setEnabled(!m_appSettings->getValue(PIN_CONNECTION_BAR, &tmpBool));

    if (!m_pCurrentFarm->isConnectionBar())
    {
        pinConnectionBarEnabled(false);
    }
}

void DisplayPage::saveFarmSettings()
{
    // Farm doesn`t exist
    if (m_pCurrentFarm == NULL)
    {
        return;
    }

    int width = 0;
    int height = 0;
    QString sizeString = desktopSizeLabel->text();
    int sizeDelimiterIndex = sizeString.indexOf("x");
    if (sizeDelimiterIndex > 0)
    {
        width = sizeString.left(sizeDelimiterIndex).toInt();
        height = sizeString.right(sizeString.length() - sizeDelimiterIndex - 1).toInt();
    }
    m_pCurrentFarm->setDesktopWidth(width);
    m_pCurrentFarm->setDesktopHeight(height);

    int colorDepth = m_colorDepthList.at(colorDepthComboBox->currentIndex());
    m_pCurrentFarm->setColorDepth(colorDepth);

    m_pCurrentFarm->setSeamlessMode(seamlessModeCheckBox->isChecked());
    m_pCurrentFarm->setConnectionBar(displayConnectionBarCheckBox->isChecked());
    // m_pCurrentFarm->setPinConnectionBar(pinConnectionBarCheckBox->isChecked());
    m_pCurrentFarm->setSpanMonitors( multimonCheckBox->isChecked());
}

void DisplayPage::displaySizeChanged(int index)
{
    desktopSizeLabel->setText(m_desktopSizeList.value(index));
}

void DisplayPage::colorDepthChanged(int index)
{
    // If index isn`t correct
    if (index < 0)
    {
        return;
    }

    QString currentPixmapName = m_pixmapFileNames.at(index);
    // No pixmap file for this index
    if (currentPixmapName.isEmpty())
    {
        return;
    }

    // Display correct pixamp on the page
    colorLabel->setPixmap(QPixmap(currentPixmapName));
}

void DisplayPage::pinConnectionBarEnabled(bool /*pinEnabled*/)
{
    // pinConnectionBarCheckBox->setEnabled(pinEnabled);
    // pinConnectionBarCheckBox->setChecked(false);
}
