// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "appPortalConfigParser.h"

#include "farm.h"

#include <QByteArray>
#include <QDomDocument>
#include <QDomNode>
#include <QString>


AppPortalConfigParser::AppPortalConfigParser(QObject* parent)
    : QObject(parent)
{
}

bool AppPortalConfigParser::parseFromBuffer(Farm* pFarm, const QByteArray& buffer)
{
    QDomDocument xmlDoc;

    if (xmlDoc.setContent(buffer))
    {
        QDomElement firstElement = xmlDoc.documentElement();
        pFarm->removeAllLocations();
        parseNextNode(pFarm, firstElement);
        pFarm->setLastLocation(0);
        return true;
    }

    return false;
}

void AppPortalConfigParser::parseNextNode(Farm* pFarm, const QDomNode& node)
{
    QDomNode domNode = node.firstChild();

    while (!domNode.isNull())
    {
        if (domNode.isElement())
        {
            QDomElement domElement = domNode.toElement();

            if (!domElement.isNull())
            {
                // If node is farm
                if (domElement.tagName() == "Farm")
                {
                    // Farm name
                    QString name = domElement.attribute("FarmName");
                    if (!name.isEmpty())
                        pFarm->setName(name);

                    parsePasswordManagementOptions(domElement, pFarm);
                    parseAdditionalOptions(domElement, pFarm);
                    parseCredentialOptions(domElement, pFarm);
                    parseLocationsOptions(domElement, pFarm);
                    parseDesktopOptions(domElement, pFarm);
                    parseRedirectionOptions(domElement, pFarm);
                    parseUserExperienceOptions(domElement, pFarm);
                }
                // If node is location
                if (domElement.tagName() == "Location")
                {
                    parseLocation(domElement, pFarm);
                }
            }// if (!domElement.isNull())

        }// if (domNode.isElement())
        parseNextNode(pFarm, domNode);
        domNode = domNode.nextSibling();
    }// while (!domNode.isNull())
}

void AppPortalConfigParser::parsePasswordManagementOptions(QDomElement& domElement, Farm* pFarm)
{
    // Password management
    QString pmServer = domElement.attribute("PasswordManagementServer");
    pFarm->setPasswordManagementServer(pmServer);
    QString spmPort = domElement.attribute("PasswordManagementPort", "443");
    int ipmPort = spmPort.toInt();
    pFarm->setPasswordManagementPort(ipmPort);
    bool pmAllow = domElement.attribute("AllowPasswordManagement", "0").toInt();
    pFarm->setPasswordManagementAllow(pmAllow);
}

void AppPortalConfigParser::parseAdditionalOptions(QDomElement& domElement, Farm* pFarm)
{
    bool connectionBar = domElement.attribute("DisplayConnectionBar", "0").toInt();
    pFarm->setConnectionBar(connectionBar);
    bool pinConnectionBar = domElement.attribute("PinConnectionBar", "0").toInt();
    pFarm->setPinConnectionBar(pinConnectionBar);
    bool seamlessMode = domElement.attribute("SeamlessMode", "0").toInt();
    pFarm->setSeamlessMode(seamlessMode);
    bool hideSettings = domElement.attribute("HideSettings", "0").toInt();
    pFarm->setHideSettings(hideSettings);
}

void AppPortalConfigParser::parseCredentialOptions(QDomElement& domElement, Farm* pFarm)
{
    // Credentials options
    bool ssoEnabled = domElement.attribute("EnableSSO", "0").toInt();
    pFarm->setSsoEnable(ssoEnabled);
    bool kerberos = domElement.attribute("EnableKerberos", "0").toInt();
    pFarm->setKerberosEnable(kerberos);
    bool kerberosMode = domElement.attribute("KerberosMode", "0").toInt();
    pFarm->setKerberosMode(kerberosMode);
    bool allowSave = !domElement.attribute("DisallowSaveCredentials", "0").toInt();
    pFarm->setCredentialsSave(allowSave);
}

void AppPortalConfigParser::parseLocationsOptions(QDomElement& domElement, Farm* pFarm)
{
    // Location options
    bool promptLocation = domElement.attribute("PromptForLocation").toInt();
    pFarm->setPromptLocation(promptLocation);
    int defaultLocation = domElement.attribute("DefaultLocation").toInt();
    pFarm->setDefaultLocation(defaultLocation);
}

void AppPortalConfigParser::parseDesktopOptions(QDomElement& domElement, Farm* pFarm)
{
    // Desktop options
    int desktopWidth = domElement.attribute("DesktopWidth", "0").toInt();
    pFarm->setDesktopWidth(desktopWidth);
    int desktopHeight = domElement.attribute("DesktopHeight", "0").toInt();
    pFarm->setDesktopHeight(desktopHeight);
    int colorDepth = domElement.attribute("ColorDepth", "0").toInt();
    pFarm->setColorDepth(colorDepth);
    bool spanMonitors = domElement.attribute("FullScreen", "0").toInt();
    pFarm->setSpanMonitors(spanMonitors);
    bool smartSizing = domElement.attribute("EnableSmartSizing", "0").toInt();
    pFarm->setSmartSizing(smartSizing);
}

void AppPortalConfigParser::parseRedirectionOptions(QDomElement& domElement, Farm* pFarm)
{
    // Sound mode switching
    int sound = domElement.attribute("AudioMode", "0").toInt();
    QString soundMode;
    switch (sound)
    {
    case 0:
        soundMode = "local";
        break;
    case 1:
        soundMode = "remote";
        break;
    default:
        soundMode = "off";
    }
    pFarm->setSoundMode(soundMode);

    // Redirection options
    QString rDrives = domElement.attribute("RedirectDrives", "0");
    pFarm->setRedirectDrives(rDrives.toInt());
    QString rPrinters = domElement.attribute("RedirectPrinters", "0");
    pFarm->setRedirectPrinters(rPrinters.toInt());
    QString rUPrinters = domElement.attribute("RedirectUniversalPrinters", "0");
    pFarm->setRedirectUPrinters(rUPrinters.toInt());
    QString rCom = domElement.attribute("RedirectComPorts", "0");
    pFarm->setRedirectComPorts(rCom.toInt());
    QString rCards = domElement.attribute("RedirectSmartCards", "0");
    pFarm->setRedirectSmartCards(rCards.toInt());
    QString rHandhelds = domElement.attribute("RedirectHandhelds", "0");
    pFarm->setRedirectHandhelds(rHandhelds.toInt());
    QString rClipboard = domElement.attribute("RedirectClipboard", "0");
    pFarm->setRedirectClipboard(rClipboard.toInt());
    QString rMicrophone = domElement.attribute("RedirectMicrophone", "0");
    pFarm->setRedirectMicrophone(rMicrophone.toInt());
}

void AppPortalConfigParser::parseUserExperienceOptions(QDomElement& domElement, Farm* pFarm)
{
    // User experience
    QString strWallpaper = domElement.attribute("EnableWallpaper", "0");
    pFarm->setWallpaperEnable(strWallpaper.toInt());
    QString strWindowDrag = domElement.attribute("EnableFullWindowDrag", "0");
    pFarm->setDraggingEnable(strWindowDrag.toInt());
    QString strAnimation = domElement.attribute("EnableAnimation", "0");
    pFarm->setAnimationEnable(strAnimation.toInt());
    QString strThemes = domElement.attribute("EnableThemes", "0");
    pFarm->setThemesEnable(strThemes.toInt());
    QString strBitmapCaching = domElement.attribute("EnableBitmapCaching", "0");
    pFarm->setBitmapCachingEnable(strBitmapCaching.toInt());
    QString reconnect = domElement.attribute("AutoReconnect", "0");
    pFarm->setReconnection(reconnect.toInt());
}

void AppPortalConfigParser::parseLocation(QDomElement& domElement, Farm* pFarm)
{
    LocationSharedPointer location(new Location);

    QString name = domElement.attribute("Name");
    if (!name.isEmpty())
        location->setName(name);

    // Setting the protocol port server
    int iProtocol = domElement.attribute("Protocol", "1").toInt();
    QString strProtocol = "https";
    if (iProtocol == 0)
        strProtocol = "http";
    location->setProtocol(strProtocol);

    QString port = domElement.attribute("TCPPort", "443");
    location->setPort(port);

    QString servers = domElement.attribute("ServerList");
    location->setServers(servers.split(","));

    // Setting ssl and nat
    bool ssl = domElement.attribute("RDPonSSL", "0").toInt();
    location->setSsl(ssl);
    bool nat = domElement.attribute("EnableNAT", "0").toInt();
    location->setNat(nat);
    QString gateway = domElement.attribute("SSLGateway");
    location->setSslGateway(gateway);

    // Proxy options
    QString proxyServer = domElement.attribute("ProxyServer");
    location->setProxyServer(proxyServer);
    QString proxyBypassList = domElement.attribute("ProxyBypassList");
    location->setProxyBypassList(proxyBypassList);

    pFarm->addLocation(location);
}
