// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <unistd.h>
#include <netdb.h>

#include <QAction>
#include <QComboBox>
#include <QtGui>
#include <QtDebug>
#include <syslog.h>
#include <QRgb>
#include <QImage>
#include <QStandardItemModel>
#include <QStandardItem>
#include <QPoint>
#include <QPushButton>

#include "mainWindow.h"
#include "ui_mainWindow.h"

#include "ManageConnectionDialog/manageConnectionDialog.h"
#include "AppSettingsDialog/appsettingsdialog.h"
#include "common/applicationSettings.h"
#include "appPortalConfigCollector.h"
#include "appPortalConfigParser.h"
#include "applicationListModel.h"
#include "FarmClient/application.h"
#include "../common/common.h"
#include "minimalviewdialog.h"
#include "UsbrDialog/usbrdialog.h"

#define APPLICATION_NAME tr("vWorkspace")
#define FARMS_CONFIG_LOCATIONS "config.xml locations"



MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent),
    settings("Quest", APPLICATION_NAME),
    m_isRemoveFarms(false),
    m_isFarmLoaded(false),
    m_requestsLeftCount(0),
    ui(new Ui::qVwWindow)
{
    currentConnection = new FarmConnection(this);

    connect(currentConnection, SIGNAL(finished(FarmConnection::FinishCode)),
            SLOT(connectionFinished(FarmConnection::FinishCode)));
    connect(currentConnection, SIGNAL(error()), SLOT(connectionError()));
    connect(currentConnection, SIGNAL(startApplication(const QString &)),
            SLOT(runApplication(const QString &)));
    connect(currentConnection, SIGNAL(changePasswordFinished(FarmConnection::FinishCode, FarmConnection*)),
            SLOT(connectionFinishedAlt(FarmConnection::FinishCode, FarmConnection*)));

    // Progress dialog
    progressDialog = new ProgressDialog( this );
    progressDialog->setWindowTitle(tr("Connection..."));
    progressDialog->setWindowModality(Qt::WindowModal);
    connect(progressDialog, SIGNAL(canceled()),
            currentConnection, SLOT(reject()));
    connect(currentConnection, SIGNAL(started()), progressDialog, SLOT(start()));

    // Create the minimal view dialog
    minimalView = new MinimalViewDialog( this );

    ui->setupUi(this);

    if ( getShowStartupTooltips() )
    {
        ui->clickHereLabel->show();
        ui->cancelClickHereButton->show();
    }

    setWindowFlags(Qt::FramelessWindowHint);

    // Connect the incoming update thumbnail request to the handler
    connect( & m_updateServer, SIGNAL( updateThumbnail(QPixmap, QString, struct pixel_update_header * )), this, SLOT( updateThumbnail( QPixmap, QString, struct pixel_update_header * )));

    // Connect the selection action of the Session Thumbnail list
    // to the handler for it.
    connect(ui->sessionListView, SIGNAL(clicked( const QModelIndex )), this, SLOT(handleSessionSelected( QModelIndex )));

    // Connect the remove session request from the minimal view
    connect( minimalView, SIGNAL(removeSessionClicked(QString)), this, SLOT(minimalViewRemoveSessionClicked(QString)));

    // Connect the refresh request from the minimal View
    connect( minimalView, SIGNAL(minimalRefreshButtonClicked()), this, SLOT(handleMinimalRefreshRequest()));

    // Connect the farm acces button from the minimal View
    connect( minimalView, SIGNAL(minimalFarmMenuButtonClicked()), this, SLOT(on_farmAccessPushbutton_clicked()));

    // Connect the minimal application list to the main model launcher
    connect( minimalView, SIGNAL(minimalViewApplicationListViewActivated(QModelIndex)),
             this, SLOT(handleMinimalViewApplicationStart(QModelIndex)));

    // Connect the minimal view session activateion requests
    connect( minimalView, SIGNAL(minimalViewActivateSession(QString)), this, SLOT(handleMinimalViewActivateSession(QString)));

    createFarmMenu();

    // Connect the models to the list views for thumbnails
    ui->sessionListView->setModel( &sessionListModel );
    // minimalView->setSessionListModel( &sessionListModel );

    // m_updateServer.setThumbnailWidth( ui->sessionListView->gridSize().width());
    m_updateServer.setThumbnailWidth( 180 );

    // Disable the refresh button to prevent crashes.
    minimalView->refreshButtonEnable(false);
    ui->refreshAppsButton->setEnabled( false );
    ui->changeUserButton->setEnabled( false );

    readSettings();
    loadFarmSettingsIfNeeded();
    updateFarmsMenu();

    // Set the logging levels
    //  Deal with the Logging settings
    if ( this->getLogging() )
    {
        setlogmask( LOG_UPTO( LOG_INFO ));
    }
    else
    {
        setlogmask( LOG_UPTO( LOG_ERR ));
    }

    statusBar()->show();

    // syslog( LOG_INFO, "Farm Menu has %i children", farmMenu->children().count());

    // This area was hidden to prevent some issues (unknown for now).
    // Now area is shown to discover is there still issue, and if it is - what exactly is this issue
    // ui->verticalFrame->hide();

}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::setRemoveFarms(bool isRemoveFarms)
{
    m_isRemoveFarms = isRemoveFarms;
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::showStartupTooltips()
{
    Farm * farm;

    if ( getShowStartupTooltips() )
    {
        // Hide the click here label
        ui->clickHereLabel->hide();
        ui->cancelClickHereButton->hide();

        int result;
        TransparentDialog w;
        w.setParent( this );
        w.move( QPoint( 0,0));
        w.resize( geometry().size());
        result = w.exec();

        // If the checkbox was checked,
        if ( result ) setShowStartupTooltips( false );

        // Show the click here label if necessary
        if ( getShowStartupTooltips() )
        {
            ui->clickHereLabel->show();
            ui->cancelClickHereButton->show();
        }
    }


    // Hide the 'Click here button if necessary
    if ( !(settings.value("ShowClickHere", true).toBool()) )
    {
        ui->clickHereLabel->hide();
        ui->cancelClickHereButton->hide();
    }

    // NOTE: This isn't exactly the right place but it works.
    //       So check and do the autoconnect here.
    // Go through the farms list and autoconnect the first one it finds.
    int farmCount = settings.getFarmNamesList().count();

    // Ported from the old App Portal
    for ( int i = 0; i < farmCount; i ++ )
    {
        farm = settings.getFarm( i );

        if ( farm->autoConnected() )
        {
            syslog( LOG_INFO, "Autoconnecting to %s", farm->getName().toAscii().data());
            farm->setCurrentLocation(farm->getDefaultLocation());
            currentConnection->setFarm( farm );
            currentConnection->connectTo();
        }
    }
}

void MainWindow::readSettings()
{
    m_usbAutoshare  = settings.value("UsbAutoshare", false).toBool();
    settings.loadFarms();
}

void MainWindow::setUsbAutoshare( bool share )
{
    settings.setValue( "UsbAutoshare",  share );
}

void MainWindow::setShowHelpSteps( bool share )
{
    settings.setValue( "ShowSteps",  share );
}

void MainWindow::setLogging( bool log )
{
    settings.setValue( "ApplicationLogging",  log );
}

bool MainWindow::getUsbAutoshare( )
{
    return settings.value("UsbAutoshare", false).toBool();
}

bool MainWindow::getLogging( )
{
    return settings.value("ApplicationLogging", false).toBool();
}

bool MainWindow::getShowHelpSteps( )
{
    return settings.value( "ShowSteps",  true ).toBool();
}

bool MainWindow::getShowStartupTooltips( )
{
    return settings.value("ShowStartupTooltips", true).toBool();
}

void MainWindow::setShowStartupTooltips( bool show )
{
    // If this is true, then also enable the 'show click here' setting
    // But only if it was initially false.  This is to prevent the changing
    // of the setting from simply opening the dialog and closing it.
    //
    // NOTE: This must be the first operation in the function because it depends on knowing
    //       if the tooltip value was actually changed or not.
    if ( show  && ( !getShowStartupTooltips())) settings.setValue( "ShowClickHere",  show );


    settings.setValue( "ShowStartupTooltips",  show );
}


void MainWindow::writeSettings()
{
    settings.saveFarms();
}

void MainWindow::loadFarmSettingsIfNeeded()
{
    // Search, download and load farms settings, if can find
    if (settings.getFarmNamesList().empty())
    {

        m_pConfigCollector = new AppPortalConfigCollector(this);
        m_pConfigParser = new AppPortalConfigParser(this);
        connect(m_pConfigCollector,
                SIGNAL(networkReply(QNetworkReply::NetworkError, QString, QString)),
                SLOT(networkReply(QNetworkReply::NetworkError, QString, QString)));

        // Try locations from global application config
        readFarmSettingsFromDefaultLocations();
    }
}

void MainWindow::readFarmSettingsFromDefaultLocations()
{
    QStringList farmConfigLocations;
    ApplicationSettings appSettings(APPLICATION_CONFIG_FILE_NAME);
    int size;

    // Reading farms setting locations list from application config file
    size = appSettings.getValueList(FARMS_CONFIG_LOCATIONS, &farmConfigLocations);
    m_requestsLeftCount = size;

    if (size > 0)
    {
        foreach (QString fullUrl, farmConfigLocations)
        {
            // If file with farm settings are not loaded yet
            if (!m_isFarmLoaded)
            {
                // Parsing location string for farm settings request
                QString protocol, url, fileName;
                int protocolEndIndex, fileBeginIndex;
                protocolEndIndex = fullUrl.indexOf(':');
                fileBeginIndex = fullUrl.lastIndexOf('/');
                // If location string parsing is successful
                if ((protocolEndIndex != -1) && (fileBeginIndex != -1) && (protocolEndIndex < fileBeginIndex))
                {
                    protocol = fullUrl.left(protocolEndIndex);
                    protocolEndIndex += QString("://").size();
                    url = fullUrl.mid(protocolEndIndex, fileBeginIndex - protocolEndIndex);
                    fileName = fullUrl.right(fullUrl.size() - fileBeginIndex - 1);

                    syslog(LOG_NOTICE, "Requesting %s://%s/%s", protocol.toAscii().data(),
                           url.toAscii().data(), fileName.toAscii().data());
                    // Requesting file with farm settings
                    m_pConfigCollector->loadConfigFromNetwork(protocol, url, fileName, false, "");
                }
                else
                {
                    --m_requestsLeftCount;
                }
            }
            else
            {
                --m_requestsLeftCount;
            }
        }

    }
    if (m_requestsLeftCount == 0)
    {
        readFarmSettingsFromPredefinedLocationMasks();
    }
}

void MainWindow::readFarmSettingsFromPredefinedLocationMasks()
{
    int result;
    char hostName[HOST_NAME_MAX];
    hostent *hostEntity;

    syslog( LOG_NOTICE, "No defined config files loaded, searching default paths" );

    // Getting host name
    result = gethostname(hostName, HOST_NAME_MAX);
    if (result == 0)
    {
        syslog( LOG_NOTICE, "My Hostname is: %s", hostName );

        // Getting FQDN in hostEntity->h_name by host name
        hostEntity = gethostbyname(hostName);
        if (result == 0)
        {
            QString domain(hostEntity->h_name);

            syslog( LOG_NOTICE, "My FULL Hostname is: %s", hostEntity->h_name );
            int domainBeginIndex = domain.indexOf('.');
            if (domainBeginIndex == -1)
            {
                QFile domainNameFile( "/tmp/domainname" );
                syslog ( LOG_INFO, "Dropping to default domain name discovery" );

                // I need to find the final fallback
                if (!domainNameFile.open(QIODevice::ReadOnly | QIODevice::Text)) return;

                // NOTE: Only expecting the single line containing the domain name
                QByteArray line = domainNameFile.readLine();
                syslog( LOG_INFO, "Last Resort domainname is %s", line.data());

                // INFO: There is no checking for sensibility of the domain name here
                // Append the new found domain to the hostname
                // Then get the domainBeginIndex as before.
                domain.append(".");
                domain.append( line );

                domainBeginIndex = domain.indexOf('.');
            }

            if (domainBeginIndex != -1)
            {
                // Getting domain name from FQDN
                domain = domain.right(domain.length() - domainBeginIndex - 1);

                // URL masks list
                const int urlCount = 4;
                QString protocol[urlCount] = {"http",        "http",       "https",       "https"};
                QString url[urlCount] =      {"vworkspace.", "provision.", "vworkspace.", "provision."};
                QString fileName[urlCount] = {"config.xml",  "config.xml", "config.xml",  "config.xml"};
                for (int i = 0; i < urlCount; ++i)
                {
                    // Appending domain name to URL
                    url[i] += domain;
                    syslog(LOG_NOTICE, "Requesting %s://%s/%s", protocol[i].toAscii().data(),
                           url[i].toAscii().data(), fileName[i].toAscii().data());
                    m_pConfigCollector->loadConfigFromNetwork(protocol[i], url[i], fileName[i], false, "");
                }
            }
            else
            {
                syslog( LOG_WARNING, "Unable to determine local domain.  Aborting config file search" );
            }
        }
        else
        {
            syslog( LOG_WARNING, "Unable to determine full local hostname. Aborting config file search" );
        }
    }
    else
    {
        syslog( LOG_ERR, "Unable to detemine local hostname.  Aborting config file search");
    }
}


void MainWindow::networkReply(QNetworkReply::NetworkError error, QString errorString, QString urlString)
{
    // If file with farm settings are not loaded yet
    if (!m_isFarmLoaded)
    {
        if (error == QNetworkReply::NoError)
        {
            // Getting buffer with loaded data
            SharedBuffer sharedBuffer = m_pConfigCollector->getConfigBuffer(urlString);
            if (sharedBuffer)
            {
                Farm *farm = settings.createNewFarm();
                settings.addFarm(farm);
                // Parsing data from buffer
                if (m_pConfigParser->parseFromBuffer(farm, *sharedBuffer))
                {
                    m_isFarmLoaded = true;
                    updateFarmsMenu();
                    // syslog(LOG_NOTICE, "Reply on %s: Configuration [%s] loaded successfully",
                    //       urlString.toAscii.data(), farm->getName().toAscii().data());
                }
                else
                {
                    // Removing empty farm if we can't parse it from file
                    settings.removeFarm(farm);
                    syslog(LOG_NOTICE, "Reply on %s: Error in parsing data", urlString.toAscii().data());
                }
            }
            else
            {
                syslog(LOG_NOTICE, "Reply on %s: There is no config", urlString.toAscii().data());
            }
        }
        else
        {
            syslog(LOG_NOTICE, "Reply on %s: Error occured - %s", urlString.toAscii().data(),
                   errorString.toAscii().data());
        }
    }
    // If file with farm settings are already loaded
    else
    {
        syslog(LOG_NOTICE, "Reply on %s: Already loaded from other url", urlString.toAscii().data());
    }

    --m_requestsLeftCount;
    // If we tried all default locations from global application config and could not load farm settings
    if (m_requestsLeftCount == 0 && !m_isFarmLoaded)
    {
        // Try locations from predefined location masks list
        readFarmSettingsFromPredefinedLocationMasks();
    }
}

void MainWindow::createFarmMenu()
{
    farmMenu.reset(new QMenu);
    // Test code for menu
    farmMenu->setStyleSheet( QString(
        "QMenu {"
        "margin-top: 20px;"
        "margin-left: 20px;"
        "border-top-width: 5px;"
        "border-left-width: 5px;"
        "border-right-width: 5px;"
        "border-bottom-width: 5px;"
        "border-style: solid;"
        "border-color: none;"
        "padding-left:0px;"
        "padding-right:0px;"
        "background: qlineargradient(spread:pad, x1:0.386, y1:0, x2:0.396, y2:1, stop:0.0409091 rgba(151, 151, 151, 255), stop:0.195455 rgba(46, 46, 46, 255), stop:0.495455 rgba(71, 71, 71, 255), stop:0.586364 rgba(181, 181, 181, 255), stop:0.631818 rgba(56, 56, 56, 255), stop:1 rgba(68, 68, 68, 255));"
        "}"


        "QPushButton {"
        "text-align: left;"
        "min-height: 1.7em;"
        "border:none;"
        "}"

        "QPushButton:hover {"
        "border-width: 1px;"
        "border-color: rgba( 0,0,0,30);"
        "border-style: solid;"
        "border-radius:3;"
        "background: rgba( 0,0,0,20); }"

        "QPushButton:pressed {"
        "background: blue; }"

        "QFrame {"
        "padding: 3px;"
        "background: white;"
        "border-bottom-width: 1px;"
        "border-bottom-color: darkgrey;"
        "border-bottom-style: solid; }"

     ));

    updateFarmsMenu();
}

void MainWindow::updateFarmsMenu()
{
    QWidgetAction * widgetAction;
    QPushButton * pb;

    syslog( LOG_INFO, "Rebuilding Farm Menu");

    // Clear and delete old menu to free resources
    // If there is a menu widget, delete all its entries.
    farmMenu->clear();

    QStringList farmNames = settings.getFarmNamesList();

    // Enable work buttons
    // put the top label

    int index = 0;

    foreach (QString name, farmNames)
    {
        index ++;    // NOTE: This counter is 1 based
        makeFarmMenuButton( name, farmMenu.get(), farmNames.count(), index );
    }

    // Make a button.
    pb = new QPushButton(" Add Farm...");
    pb->setFlat( true );
    pb->setStyleSheet( QString(
           "QPushButton {"
           "max-width: 200px;"
           "border-top-style: double;"
           "border-top-color: darkgrey;"
           "border-top-width: 2px;"
           "border-bottom-left-radius: 5px;"
           "border-bottom-right-radius: 5px;"
           "background-color: white;"
           "padding: 3px;"
           "text-align: left;}"
           ));
    widgetAction = new QWidgetAction(this);
    widgetAction->setDefaultWidget(pb);
    widgetAction->setObjectName( QString("AddFarm"));
    farmMenu->addAction(widgetAction);
    connect( pb, SIGNAL(clicked( bool )), this, SLOT( handleAddFarm( bool )));

    farmMenu->resize( farmMenu->sizeHint());
    QPixmap *mask = MaskFactory::createTopPointerMask( farmMenu->width(), farmMenu->height(), 7 );
    farmMenu->setMask( mask->createMaskFromColor( QColor( "black"), Qt::MaskOutColor ));
}





//-----------------------------------------------------------------------------
//
//  Farm Application List Handling
//
//-----------------------------------------------------------------------------

// Handle the request coming in from the minimal view
void MainWindow::handleMinimalRefreshRequest( void )
{

        syslog( LOG_INFO, "Received signal from minimal view to refresh a farm");
        syslog( LOG_INFO, "Moving progress dialog to: %ix%i",
                minimalView->x() + minimalView->width()/2 - progressDialog->width()/2 -150,
                minimalView->y() + minimalView->height()/2 - progressDialog->height()/2 );

        progressDialog->move( QPoint( minimalView->x() + minimalView->width()/2 - progressDialog->width()/2,
                                      minimalView->y() + minimalView->height()/2 - progressDialog->height()/2 ));

        refreshApplicationSet();
}

void MainWindow::refreshApplicationSet()
{
    QString processName = tr("Refreshing an application list...");

    currentConnection->refreshApplicationsList();
}

void MainWindow::runFarm(const QString & currentFarmName)
{
    // FIXME: There can be name collisions here.  This must be
    //        modified to run against the farm key or index number
    //        in the farm list.

    // Initialize farm connection
    Farm *currentFarm = settings.getFarm(currentFarmName);
    currentConnection->setFarm(currentFarm);

    syslog( LOG_INFO, "Attepting to load apps for farm %s", currentFarmName.toAscii().data());

    if ( currentFarmName.isEmpty() ) return;

    // this->treeMenuWidget->clear();

    QString processName = tr("Connecting to farm...");
    // this->setStatus(processName);
    bool result  = false;

    do
    {
        result = currentConnection->connectTo();
        if (currentConnection->lastError().contains(tr("Logon failure")))
        {
            currentConnection->setRequestCredentials(true);
        }

    } while(currentConnection->lastError().contains(tr("Logon failure")));

    // Some troubles
    if (!result)
    {
        QString lastError = currentConnection->lastError();
        if (!lastError.isEmpty())
        {
            QMessageBox::information(this,
                                     tr("Connection error..."),
                                     lastError);
        }
    }
    this->updateFarmsMenu();
}


void MainWindow::updateLocationLabel( QString folder )
{
    QString newConnectionInfo;
    QString locationName, farmName;

    syslog( LOG_INFO, "Setting folder to: %s", folder.toAscii().data());

    newConnectionInfo.append( QString("<html>Applications [") );

    farmName = currentConnection->farm()->getName();
    farmName.replace( " ", "&nbsp;");
    newConnectionInfo.append( farmName );
    newConnectionInfo.append( " : " );

    locationName =  currentConnection->farm()->getCurrentLocation()->getName();
    locationName.replace( " ", "&nbsp;");
    newConnectionInfo.append( locationName );
    newConnectionInfo.append( "]" );

    folder.replace( " ", "&nbsp;");
    if ( !folder.isEmpty()) newConnectionInfo.append( " >&nbsp;" + folder + "</html>" );
    ui->applicationGroupLabel->setText( newConnectionInfo );

    newConnectionInfo = QString("<html>[");
    newConnectionInfo.append( currentConnection->farm()->getName());
    newConnectionInfo.append( " : " );
    newConnectionInfo.append( currentConnection->farm()->getCurrentLocation()->getName());
    newConnectionInfo.append( "]" );
    if ( !folder.isEmpty()) newConnectionInfo.append( "<br>> " + folder + "</html>");
    minimalView->setMinimalViewFarmLocationLabel( newConnectionInfo );

}

void MainWindow::connectionFinished(FarmConnection::FinishCode exitCode)
{
    BasicMessageDialog messageBox;

    // Updating status bar message
    progressDialog->hide();

    switch (exitCode)
    {
    case FarmConnection::Success:
        // syslog( LOG_INFO, "Sucessfully retrieved the application list");

        // Load the model into the list view
        appListModel.setApplicationList(currentConnection->applications());

        // Assign the model to the list view
        ui->applicationListView->setModel( &appListModel );
        ui->applicationListView->update();

        // Assign the model to the minimal view's list
        minimalView->setApplicationListModel( &appListModel );

        // Update the various UI elements
        updateLocationLabel( "" );

        // Hide the 'Click here' label
        ui->clickHereLabel->hide();
        ui->cancelClickHereButton->hide();

        // Enable the refresh button
        minimalView->refreshButtonEnable(true);
        ui->refreshAppsButton->setEnabled( true );
        ui->changeUserButton->setEnabled( true );

        break;

    case FarmConnection::Error:
        // Connection was failed

        // don't do this on launching an application
        if (!currentConnection->isConnected())
        {
            // treeMenuWidget->clear();

            // Disabling options for our connection
            // refreshApplicationSetAction->setEnabled(false);
            // changeCurrentConnectionAction->setEnabled(false);
            // logonAsADifferentUserAction->setEnabled(false);
            minimalView->refreshButtonEnable(false);
            ui->refreshAppsButton->setEnabled( false );
            ui->changeUserButton->setEnabled( false );

        }

        // Displaying last connection error
        syslog( LOG_ERR, "Error Message: %s", currentConnection->lastError().toAscii().data());
        messageBox.initialize( tr("Connection error..."), currentConnection->lastError() );
        messageBox.exec();
        break;

    case FarmConnection::Reject:
        syslog( LOG_INFO, "Request to retrieve the application list was rejected");
        // Connection was rejected

        break;
    case FarmConnection::Message:
        // Need to show the message
        QString message = currentConnection->lastError();
        syslog( LOG_INFO, "Received Message: %s", message.toAscii().data());

        messageBox.initialize( tr("Changing password"), message );
        messageBox.exec();
        break;
    }
}

void MainWindow::connectionError()
{
        // Updating status bar message
        statusBar()->showMessage(tr("Connection failed"));

        // Displaying last connection error
        // QString lastError = currentConnection->lastError();
        // QMessageBox::critical(this, tr("Connection error..."), lastError);

        // Clearing app list
        // treeMenuWidget->setVisible(false);
        // treeMenuWidget->clear();

        // Disabling options for our connection
        // refreshApplicationSetAction->setEnabled(false);
        // changeCurrentConnectionAction->setEnabled(false);
        // logonAsADifferentUserAction->setEnabled(false);
}

void MainWindow::runApplication(const QString & appName)
{
    Farm * currentFarm = currentConnection->farm();
    if (currentFarm == 0)
        return;

    ApplicationList appList = currentFarm->applications();
    if (appList.isEmpty())
        return;

    Application * app = appList.application(appName);
    if (app == 0)
        return;


    currentConnection->executeSelectedApplication(appName);
}


void MainWindow::connectionFinishedAlt(FarmConnection::FinishCode exitCode, FarmConnection *connection)
{
    BasicMessageDialog messageBox;

    progressDialog->hide();

    switch (exitCode)
    {
    case FarmConnection::Success:
        messageBox.initialize(tr("Password Management..."), tr("Password was changed successfully."));
        messageBox.exec();
        break;
    case FarmConnection::Error:
        // Displaying last connection error
        messageBox.initialize(tr("Changing password error..."), connection->lastError());
        messageBox.exec();
        break;
    case FarmConnection::Reject:
        break;
    case FarmConnection::Message:
        // Need to show the message
        messageBox.initialize(tr("Changing password"), connection->lastError());
        messageBox.exec();
        break;
    }
}

void MainWindow::handleAddFarm( bool checked )
{
    QPoint  local_pos;
    QWidget * childWidget ;
    SimpleAddFarmDialog * dialog;

    // Clearing the warning
    if (checked) {};

    // syslog( LOG_INFO, "A farm add is requested");

    local_pos = farmMenu->mapFromGlobal( QCursor::pos());
    childWidget = farmMenu->childAt( local_pos );

    // syslog( LOG_INFO, "The keypress was detected at location %ix%i on the widget.", local_pos.x(), local_pos.y());
    // syslog ( LOG_INFO, "The menu's child widget is a %s", childWidget->objectName().toAscii().data());
    QPushButton * pb = (QPushButton *) sender();
    pb->clearFocus();
    farmMenu->close();

    // VwMenuDialog dialog;
    dialog = new SimpleAddFarmDialog( &settings, NULL );

    dialog->move( QPoint(( x() + width()/2) - dialog->width()/2,
                        ( y() + height()/2) - dialog->height()/2 ));

    dialog->exec();
    //dialog->close();
}



void MainWindow::handleFarmMenuMouseEnter( QString name )
{
    QString farmName;
    QStringList locationList;
    Farm * farm;
    int count;

    // If there is a previous location menu open,
    // close it and open the new one.
    if ( locationMenu != NULL )
    {
        locationMenu->close();
        locationMenu = NULL;
    }

    // Check to see if there is a default location set
    // If there isn't one, then pop up the location dialog
    // to allow the user to choose which location to activate, and
    // also allow the setting of the default buttons
    // Get the farm name, and then the farm
    farmName = name;
    farm = settings.getFarm( farmName );

    if ( farm->isPromptLocation() )
    {
        locationMenu = new QMenu;

        locationMenu->setStyleSheet( QString(

        "QMenu {"
        "margin-top: 20px;"
                "margin-left: 20px;"
                "border-top-width: 5px;"
                "border-left-width: 5px;"
                "border-right-width: 5px;"
                "border-bottom-width: 5px;"
                "border-style: solid;"
                "border-color: none;"
                "padding-left:0px;"
                "padding-right:0px;"
                "background: qlineargradient(spread:pad, x1:0, y1:0.472, x2:1, y2:0.466, stop:0.0409091 rgba(151, 151, 151, 255), stop:0.195455 rgba(46, 46, 46, 255), stop:0.495455 rgba(71, 71, 71, 255), stop:0.586364 rgba(181, 181, 181, 255), stop:0.631818 rgba(56, 56, 56, 255), stop:1 rgba(68, 68, 68, 255));"
                "}"

        "QPushButton {"
        "text-align: left;"
        "min-height: 1.7em;"
        "border:none;"
        "}"

        "QPushButton:hover {"
        "border-width: 1px;"
        "border-color: rgba( 0,0,0,30);"
        "border-style: solid;"
        "border-radius:3;"
        "background: rgba( 0,0,0,20); }"

        "QPushButton:pressed {"
        "background: blue; }"

        "QFrame {"
        "padding: 3px;"
        "background: white;"
        "border-bottom-width: 1px;"
        "border-bottom-color: darkgrey;"
        "border-bottom-style: solid; }"
         ));

        locationList = farm->getLocationNamesList();
        count = locationList.count();
        int index = 0;

         // For each location make a button
        foreach (QString name, locationList)
        {
            index ++;   // Note that this counter is '1' based.
            makeLocationMenuButton( name, farmName, locationMenu, count, index );
        }

        // Put the location menu where it belongs
        locationMenu->move( (farmMenu->x()+farmMenu->width())-20, QCursor::pos().y()-40);
        locationMenu->show();
    }
    /*
    else
    {
        // If a default location is set, then just run the farm.
        // get the default location and set it to the current
        farm->setCurrentLocation(farm->getDefaultLocation());
        runFarm( farmName );
        farmMenu->close();
    }
    */

}

void MainWindow::handleFarmMenuMouseLeave( QString name )
{
    if ( locationMenu != NULL )
    {
        locationMenu->close();
        // delete locationMenu;
        locationMenu = NULL;
    }
}




void MainWindow::handleFarmMenuClick( bool checked )
{
    QString farmName;
    QStringList locationList;
    Farm * farm;
    int count;

    // Clearing the warning
    if (checked) {};

    QPushButton * pb = (QPushButton *) sender();
    pb->clearFocus();

    // Check to see if there is a default location set
    // If there isn't one, then pop up the location dialog
    // to allow the user to choose which location to activate, and
    // also allow the setting of the default buttons
    // Get the farm name, and then the farm
    farmName = sender()->objectName();
    farm = settings.getFarm( farmName );

    if ( farm->isPromptLocation() )
    {
        // If there is a previous location menu open,
        // close it and open the new one.
        if ( locationMenu != NULL )
        {
            locationMenu->close();
            locationMenu = NULL;
        }

        locationMenu = new QMenu;

        locationMenu->setStyleSheet( QString(

        "QMenu {"
        "margin-top: 20px;"
                "margin-left: 20px;"
                "border-top-width: 5px;"
                "border-left-width: 5px;"
                "border-right-width: 5px;"
                "border-bottom-width: 5px;"
                "border-style: solid;"
                "border-color: none;"
                "padding-left:0px;"
                "padding-right:0px;"
                "background: qlineargradient(spread:pad, x1:0, y1:0.472, x2:1, y2:0.466, stop:0.0409091 rgba(151, 151, 151, 255), stop:0.195455 rgba(46, 46, 46, 255), stop:0.495455 rgba(71, 71, 71, 255), stop:0.586364 rgba(181, 181, 181, 255), stop:0.631818 rgba(56, 56, 56, 255), stop:1 rgba(68, 68, 68, 255));"
                "}"

        "QPushButton {"
        "text-align: left;"
        "min-height: 1.7em;"
        "border:none;"
        "}"

        "QPushButton:hover {"
        "border-width: 1px;"
        "border-color: rgba( 0,0,0,30);"
        "border-style: solid;"
        "border-radius:3;"
        "background: rgba( 0,0,0,20); }"

        "QPushButton:pressed {"
        "background: blue; }"

        "QFrame {"
        "padding: 3px;"
        "background: white;"
        "border-bottom-width: 1px;"
        "border-bottom-color: darkgrey;"
        "border-bottom-style: solid; }"
         ));

        locationList = farm->getLocationNamesList();
        count = locationList.count();
        int index = 0;

         // For each location make a button
        foreach (QString name, locationList)
        {
            index ++;   // Note that this counter is '1' based.
            makeLocationMenuButton( name, farmName, locationMenu, count, index );
        }

        // Put the location menu where it belongs
        locationMenu->move( (farmMenu->x()+farmMenu->width())-20, QCursor::pos().y()-40);
        locationMenu->show();

    }
    else
    {
        // If a default location is set, then just run the farm.
        // get the default location and set it to the current
        farm->setCurrentLocation(farm->getDefaultLocation());
        runFarm( farmName );
        farmMenu->close();
    }
}

void MainWindow::handleFarmMenuSettingsClick( bool checked )
{
    // syslog ( LOG_INFO, "'%s' settings button was pressed", this->sender()->objectName().toAscii().data());

    // Clearing the warning
    if (checked) {};

    ManageConnectionDialog *connectionDialog = new ManageConnectionDialog(&settings, false, sender()->objectName(), NULL);

    // Farm was changed
    if (connectionDialog->exec() == QDialog::Accepted && currentConnection != NULL)
    {
        // this->updateFarmsList();
        Farm *currentFarm = currentConnection->farm();
        if (currentFarm != NULL && currentFarm->needsReconnecting())
        {
            // int row = settings.getFarmNamesList().indexOf(currentFarm->getName());
            // farmListWidget->setCurrentRow(row);
            runFarm( currentFarm->getName() );
            currentFarm->setNeedsReconnecting(false);
        }
    }

    QPushButton * pb = (QPushButton *) sender();
    pb->clearFocus();
    farmMenu->close();

    updateFarmsMenu();

}


// This is the common handler for the location submenu.  If the 'default' button is pressed,
// then the 'makeDefault' bool is 'true' and the value will be saved.
void MainWindow::handleLocationSelectionButton( QString buttonName, bool makeDefault )
{
    QString farmName, locationName;
    Farm * farm;
    int markerIndex;

    // Get the farm name, and then the farm
    markerIndex = buttonName.indexOf( QString( "_***_" ));
    farmName = buttonName.left( markerIndex );
    farm = settings.getFarm( settings.getFarmNamesList().indexOf( farmName ));

    // Get the location name and then the location index
    locationName = buttonName.right( buttonName.length() - ( 5 + farmName.length()) );

    syslog( LOG_INFO, "Location handler found: Farm:%s and Location:%s.", farmName.toAscii().data(), locationName.toAscii().data());

    // Get the location index from the farm's location list
    farm->setCurrentLocation( farm->getLocationNamesList().indexOf( locationName ));

    if ( makeDefault )
    {
        // Also set this as the default now if told to
        farm->setDefaultLocation( farm->getLocationNamesList().indexOf( locationName ) );

        // Turn off the location prompt
        farm->setPromptLocation( false );
    }

    QPushButton * pb = (QPushButton *) sender();
    pb->clearFocus();

    // Load the farm
    runFarm( farmName );

    return;
}

void MainWindow::handleLocationClick( bool checked  )
{
    // Clearing the warning
    if (checked) {};

    QPushButton * pb = (QPushButton *) sender();
    pb->clearFocus();

    handleLocationSelectionButton( sender()->objectName(), false );

    return;
}

void MainWindow::handleLocationDefaultClick( bool checked )
{
    // Clearing the warning
    if (checked) {};

    QPushButton * pb = (QPushButton *) sender();
    pb->clearFocus();

    handleLocationSelectionButton( sender()->objectName(), true );

    return;
}

void MainWindow::handleMinimalViewApplicationStart( QModelIndex index )
{
    on_applicationListView_doubleClicked( index );
}


void MainWindow::handleSessionSelected( QModelIndex index )
{
    QPoint  local_pos, child_pos;
    QWidget * childWidget ;
    QStandardItem * item;
    int wid, seamless;

    // Clear the selection
    ui->sessionListView->clearSelection();

    item = sessionListModel.item( index.row(), 0 );
    wid      = item->data( Qt::UserRole + 1 ).toUInt();
    seamless = item->data( Qt::UserRole + 2 ).toUInt();

    // FIXME: Error handling please

    syslog( LOG_INFO,"Session selected: 0x%08X", wid );

    // decide if this is in the top left corner or not.
    local_pos = ui->sessionListView->mapFromGlobal( QCursor::pos());
    childWidget = ui->sessionListView->childAt( local_pos );
    child_pos = childWidget->mapFromGlobal( QCursor::pos() );

    // Dont forget to add in the scrolled amount.
    int scroll_x = ui->sessionListView->horizontalScrollBar()->value();

    // syslog( LOG_INFO, "The keypress was detected at location %ix%i on the widget.", local_pos.x() + scroll_x, local_pos.y());
    // syslog( LOG_INFO, "The keypress was detected at location %ix%i on the child widget.", child_pos.x() + scroll_x, child_pos.y());

    // NOTE: The 220px is the grid size of the list view.

    if (QRect( 2, 16, 16, 16 ).contains( (child_pos.x()+scroll_x) % 220 , child_pos.y()))
    {
        syslog ( LOG_INFO, "A Close session request is being sent to the window");

        m_updateServer.closeWindow( wid );

        // This pause is necessary to allow the remote session to cleanly shutdown and
        // send it's last few messages.
        usleep( 1000000 );

        // Manually remove the thumbnail now
        // Use theindex to find the item
        // Remove the item
        QString sessionName;
        sessionName = sessionListModel.itemFromIndex( index )->text();
        syslog( LOG_INFO, "About remove thumbnail '%s'(%i)", sessionName.toAscii().data(), index.row() );
        sessionListModel.removeRow( index.row(), QModelIndex() );
        minimalView->removeSession( sessionName );
        ui->sessionListView->update();
        ui->sessionListView->clearSelection();

    }
    else
    {
        // Only do the raise if this is NOT a seamless session
        if ( seamless == 0 ) m_updateServer.raiseWindow( wid );
    }

    /*
    syslog( LOG_INFO, "The keypress was detected at location %ix%i on the widget.", local_pos.x(), local_pos.y());
    syslog( LOG_INFO, "The keypress was detected at location %ix%i on the child widget.", child_pos.x(), child_pos.y());
    syslog( LOG_INFO, "The list widget has %i children", ui->sessionListView->children().count());
    syslog ( LOG_INFO, "The item's widget is a %s", childWidget->objectName().toAscii().data());
    syslog( LOG_INFO, "The scroll bar is at %i.", ui->sessionListView->horizontalScrollBar()->value());
    */
}



void MainWindow::showStatusMessage(const QString & status)
{
    statusBar()->showMessage(status);

}


void MainWindow::updateThumbnail( QPixmap thumbnail , QString sessionName, struct pixel_update_header *hdr )
{
    QStandardItem * sitem;
    QList <QStandardItem *> sitems;

    // Look for a session already of the same name
    sitems = sessionListModel.findItems( sessionName, Qt::MatchFixedString | Qt::MatchCaseSensitive );

    // syslog( LOG_INFO, "Received a Thumbnail event with thumbnail dimensions %ix%i for session %s, action:%i",
    //         thumbnail.width(),
    //         thumbnail.height(),
    //         sessionName.toAscii().data(),
    //         hdr->action );

    // Seamless session updates were ignored to prevent some issues (unknown for now).
    // Now seamless session updates work to discover is there still issue, and if it is - what exactly is this issue
/*
    if ( hdr->seamless )
    {
        syslog( LOG_INFO, "Ignoring seamless session updates" );
        return;
    }
*/
    //
    // Delete a thumbnail.  This does NOT shut the session down.
    // This is in response to the session requesting a shutdown
    // at the time the message is clicked.
    //
    // If the incoming image has 0x0 dimensions, then remove it from the thumbs list.
    if ( hdr->action == AU_ACTION_REMOVE_SESSION )
    {
        syslog( LOG_INFO,"Received a remove session request, attempting to remove session thumbnail");

        // If a thumbnail already exists, then remove it
        if ((sitems.count()) > 0 )
        {
            sitem = sitems.first();

            syslog( LOG_INFO, "About to remove thumbnail index %i.", sitem->index().row());
            // Remove the item
            sessionListModel.removeRow( sitem->index().row(), QModelIndex());
            minimalView->removeSession( sessionName );
            ui->sessionListView->update();
            ui->sessionListView->clearSelection();
        }
        // Else complain a bit
        else
        {
            syslog( LOG_WARNING, "Did not find thumbnail for session %s.", sessionName.toAscii().data());
            syslog( LOG_INFO, "It might have been previously deleted");
        }
        return;
    }

    //
    // Update a thumbnail
    //
    // If a thumbnail already exists, then update it
    if ((sitems.count()) > 0 )
    {
        // Assume it is the first one, as there should only be the one.
        sitem = sitems.first();

        ui->sessionListView->setUpdatesEnabled( false );
        sitem->setIcon( QIcon( thumbnail ));
        sitem->setText( sessionName );
        ui->sessionListView->setUpdatesEnabled( true );
        ui->sessionListView->update();

    }

    //
    // Add a new thumbnail
    //
    // Else, add it
    else
    {
        // Add a new item to the main view
        sitem = new QStandardItem( QIcon( thumbnail ) , sessionName );
        sitem->setData(hdr->window_id, Qt::UserRole + 1 );
        sitem->setData(hdr->seamless, Qt::UserRole + 2 );
        sitem->setSizeHint( QSize( 180,150));
        sitem->setTextAlignment( Qt::AlignVCenter | Qt::AlignHCenter );

        sessionListModel.appendRow( sitem );
        ui->sessionListView->update();

        minimalView->addSession( sessionName );

        // syslog( LOG_INFO, "Added item with userdata 0x%08X", window_id );
    }

}



void MainWindow::minimalViewRemoveSessionClicked( QString name )
{
    int wid;
    QStandardItem * item = NULL;

    // syslog( LOG_INFO, "Received request to remove session named '%s' from minimal view", name.toAscii().data());

    // Find the session with that name
    // There should only be one
    item = sessionListModel.findItems( name, Qt::MatchExactly, 0).first();

    // Obtain the WID
    wid = item->data( Qt::UserRole + 1 ).toUInt();

    // FIXME: Error handling please

    // Shut the window down, I already know this is the correct
    // session from the name, and that my action is to shut
    // it down
    m_updateServer.closeWindow( wid );


}

void MainWindow::handleMinimalViewActivateSession( QString name )
{
    int wid;
    QStandardItem * item = NULL;

    item = sessionListModel.findItems( name, Qt::MatchExactly, 0).first();
    wid  = item->data( Qt::UserRole + 1 ).toUInt();

    m_updateServer.raiseWindow( wid );

}


void MainWindow::makeFarmMenuButton( QString name, QMenu * farmMenu, int count, int index )
{
    // Lets see what happens with a frame
    QFrame * frame1 = new QFrame;
    QHBoxLayout * frame1Layout = new QHBoxLayout;
    frame1Layout->setContentsMargins( 2, 1, 2, 1 );
    frame1Layout->setObjectName( tr("Frame's Layout"));

    // If this is the first one, then round the tops
    if ( index == 1 )
    {
        frame1->setStyleSheet( QString(
                "QFrame {"
                "max-width: 200px;"
                "min-width: 200px;"
                "border-top-left-radius: 5px;"
                "border-top-right-radius: 5px;"
                "}"
                ));
    }
    else
    {
        frame1->setStyleSheet(( QString(
                "QFrame {"
                "max-width: 200px;"
                "min-width: 200px;"
                "}"
                )));
    }


    // Make the Farm button
    MouseOverPushButton * mb = new MouseOverPushButton( name , NULL );
    mb->setObjectName( name );
    mb->setStyleSheet( QString(
           "QPushButton {"
           "padding-left: 5px;"
           "padding-right: 5px;"
           "border-radius: 5px;"
           "text-align: left;}"
                                ));
    frame1Layout->addWidget( mb );

    connect( mb, SIGNAL(clicked( bool )), this, SLOT( handleFarmMenuClick( bool )));
    connect( mb, SIGNAL(mouseEnter(QString)), this, SLOT(handleFarmMenuMouseEnter(QString)));

    // connect( mb, SIGNAL(mouseLeave(QString)), this, SLOT(handleFarmMenuMouseLeave(QString)));

    // Make the settings button
    QPushButton * pb = new QPushButton(QIcon(":/img2/Gear_Grey.png"),tr(""));
    pb->setObjectName( name );
    pb->setStyleSheet( QString(
           "QPushButton {"
           "max-width: 30px;"
           "min-width: 30px;"
           "max-height: 25px;"
           "min-height: 25px;"
           "text-align: center;"

           "border-width: 1px;"
           "border-color:darkgrey;"
           "border-style: solid;"
           "border-radius: 5px;"
           "}"

           ));
    frame1Layout->addWidget( pb );
    frame1->setLayout( frame1Layout );
    connect( pb, SIGNAL(clicked( bool )), this, SLOT( handleFarmMenuSettingsClick( bool )));

    QWidgetAction * widgetAction = new QWidgetAction(this);
    widgetAction->setDefaultWidget(frame1);
    farmMenu->addAction(widgetAction);

}

void MainWindow::makeLocationMenuButton( QString name, QString farmName, QMenu * locationMenu, int count, int index )
{
    QString buttonName;

    // Lets see what happens with a frame
    QFrame * frame1 = new QFrame;
    QHBoxLayout * frame1Layout = new QHBoxLayout;
    frame1Layout->setContentsMargins( 2, 1, 2, 1 );
    frame1Layout->setObjectName( tr("Frame's Layout"));

    // If this is the first one, then round the tops
    if ( index == 1 )
    {
        frame1->setStyleSheet( QString(
                "QFrame {"
                "border-top-left-radius: 5px;"
                "border-top-right-radius: 5px;"
                "}"
                ));
    }

    // If this is the last one, then round the bottoms
    if ( count == index )
    {
        frame1->setStyleSheet( QString(
                "QFrame {"
                "border-bottom-left-radius: 5px;"
                "border-bottom-right-radius: 5px;"
                "}"
                ));
    }

    // Make the Location button, and append the farm name
    // to the location name to make the whole button name
    QPushButton * pb = new QPushButton( name );
    buttonName = farmName;
    buttonName.append( "_***_" );
    buttonName.append( name );
    pb->setObjectName( buttonName );
    pb->setStyleSheet( QString(
           "QPushButton {"
           "padding-left: 5px;"
           "padding-right: 5px;"
           "border-radius: 5px;"
           "text-align: left;}"
                                ));
    frame1Layout->addWidget( pb );
    connect( pb, SIGNAL(clicked( bool )), this, SLOT( handleLocationClick( bool )));

    // Make the Defaulting button
    pb = new QPushButton( tr("Default") );
    pb->setObjectName( buttonName );
    pb->setStyleSheet( QString(

            "QPushButton {"
            "color: grey;"
            "padding-left: 5px;"
            "padding-right: 5px;"
            "border-radius: 5px;"
            "text-align: right;}"

           ));
    frame1Layout->addWidget( pb );
    frame1->setLayout( frame1Layout );
    connect( pb, SIGNAL(clicked( bool )), this, SLOT( handleLocationDefaultClick( bool )));

    QWidgetAction * widgetAction = new QWidgetAction(this);
    widgetAction->setDefaultWidget(frame1);
    locationMenu->addAction(widgetAction);

    locationMenu->resize( locationMenu->sizeHint());
    QPixmap *mask = MaskFactory::createLeftPointerMask( locationMenu->width(), locationMenu->height(), 7 );
    locationMenu->setMask( mask->createMaskFromColor( QColor( "black"), Qt::MaskOutColor ));

    // Hook in the buttons and reference 'ChangeLocationDialog::connectFarm()'
    // details on how to hook in the location event.

}




//////////////////////////////////////////////////////////////
//
// Mouse event capture
//
// Grab these events to make sure that they happen even though
// there may not be a widget location.  This allows the window
// to be moved even though there is no title bar.
//
//////////////////////////////////////////////////////////////
void MainWindow::mousePressEvent(QMouseEvent* event)
{
    if (event->button() == Qt::LeftButton)
    {
        m_dragPosition = event->globalPos() - frameGeometry().topLeft();
        event->accept();
    }
}

void MainWindow::mouseMoveEvent(QMouseEvent* event)
{
    if (event->buttons() & Qt::LeftButton)
    {
        move(event->globalPos() - m_dragPosition);
        event->accept();
    }
}


//////////////////////////////////////////////////////////////////
//
// Automatic Qt button connections
//
// NOTE: Thise are Qt special functions and the on_xxx_<action>()
//       naming convention has special meanings.
//
//////////////////////////////////////////////////////////////////

void MainWindow::on_mainCloseButton_clicked()
{
    syslog( LOG_WARNING, "App Portal closed from Main View");

    settings.saveFarms();
    this->close();
}

void MainWindow::on_refreshAppsButton_clicked()
{
    // Move the progress dialog to the correct spot
    progressDialog->move( QPoint( x() + width()/2 - progressDialog->width()/2,
                                  y() + height()/2 - progressDialog->height()/2 ));
    // Check that something is actually loaded before attempting a refresh
    // or else the application will crash.
    refreshApplicationSet();
}


void MainWindow::on_mainShrinkButton_clicked()
{
    // Hide this main view and unhide the minimal view
    syslog( LOG_INFO, "Switching to minimal view");
    minimalView->move(x(),y());
    minimalView->show();
    setVisible( false );


}

void MainWindow::on_applicationListView_doubleClicked(QModelIndex index)
{
    // Process item double click
    // syslog( LOG_INFO, "Activating index %i", index.row());

    // Start at the root.
    static int lastFolderIndex = -1;

    // Move the progress window
    if ( sender()->objectName() == "MinimalViewDialog")
    {
        progressDialog->move( QPoint( minimalView->x() + minimalView->width() / 2 - progressDialog->width() / 2,
                                      minimalView->y() + minimalView->height() / 2 - progressDialog->height() / 2 ));

    }
    else
    {
        progressDialog->move( QPoint( x() + width() / 2 - progressDialog->width() / 2,
                                      y() + height() / 2 - progressDialog->height() / 2 ));
    }

    // Get the type of item that I clicked on
    switch (index.data(ItemTypeRole).toInt())
    {
        case ApplicationItemType:
        {
            // Get the apps main index from the item data

            int appListIndex = index.data(ItemIndexRole).toInt();
            // syslog( LOG_INFO, "Launching application (%i)'%s'", appListIndex, currentConnection->applications().at(appListIndex)->name().toAscii().data());

            currentConnection->executeSelectedApplication(currentConnection->applications().at(appListIndex)->name());
            if (lastFolderIndex > -1)
            {
                appListModel.openFolder(lastFolderIndex);
                updateLocationLabel( appListModel.getName( lastFolderIndex ));
            }
        }
        break;
        case FolderItemType:
        {
            lastFolderIndex = index.row();
            appListModel.openFolder(lastFolderIndex);
            updateLocationLabel( appListModel.getName( lastFolderIndex ));
        }
        break;
        case BackItemType:
        {
            // Since there is only ever one level of folders, just do that.
            appListModel.showRoot();
            lastFolderIndex = -1;
            updateLocationLabel( "" );
        }
        break;
        default:
        break;
    }
}

void MainWindow::on_farmAccessPushbutton_clicked()
{
    QPoint pos = ui->farmAccessPushbutton->mapToGlobal( QPoint( 0,0));

    // Move the farmMenu then show it.
    // Put the location menu where it belongs
    syslog( LOG_INFO, "Received signal to access a farm from object named: %s", sender()->objectName().toAscii().data() );

    if ( sender()->objectName() == "MinimalViewDialog")
    {
        syslog( LOG_INFO, "Received signal from minimal view to access a farm");
        syslog( LOG_INFO, "Moving progress dialog to: %ix%i",
                minimalView->x() + minimalView->width()/2 - progressDialog->width()/2 -150,
                minimalView->y() + minimalView->height()/2 - progressDialog->height()/2 );

        progressDialog->move( QPoint( minimalView->x() + minimalView->width()/2 - progressDialog->width()/2,
                                      minimalView->y() + minimalView->height()/2 - progressDialog->height()/2 ));



    }
    else
    {
        progressDialog->move( QPoint( x() + width()/2 - progressDialog->width()/2,
                                      y() + height()/2 - progressDialog->height()/2 ));
    }

    createFarmMenu();
    // If the sender is the minimal view move the progress dialog accordingly.
    farmMenu->move( pos.x() - 20 , pos.y()+ (ui->farmAccessPushbutton->height()));
    farmMenu->show();
}

void MainWindow::on_changeUserButton_clicked()
{
    progressDialog->move( QPoint( x() + width()/2 - progressDialog->width()/2,
                                  y() + height()/2 - progressDialog->height()/2 ));
    currentConnection->logonAsADifferentUser();
}

void MainWindow::on_changePasswordButton_clicked()
{
    currentConnection->changePassword( &settings );
}

void MainWindow::on_settingsButton_clicked()
{
    AppSettingsDialog dialog;

    dialog.move( QPoint(( x() + width()/2) - dialog.width()/2,
                        ( y() + height()/2) - dialog.height()/2 ));
    dialog.exec();
}

void MainWindow::on_helpButton_clicked()
{
    MainAboutDialog dialog;

    dialog.move( QPoint(( x() + width()/2) - dialog.width()/2,
                        ( y() + height()/2) - dialog.height()/2 ));
    dialog.exec();
}


void MainWindow::on_cancelClickHereButton_clicked()
{
    // Set the internal setting for click here button
    ui->clickHereLabel->hide();
    ui->cancelClickHereButton->hide();

    // Make this permanent so it doesn't show again.
    settings.setValue( "ShowClickHere",  false );

}

void MainWindow::on_usbrButton_clicked()
{
    UsbrDialog dialog;

    dialog.move( QPoint(( x() + width()/2) - dialog.width()/2,
                        ( y() + height()/2) - dialog.height()/2 ));
    dialog.exec();

}
