// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <QtGui/QApplication>

#include "connectionbarwindow.h"
#include "qrdesktopobserver.h"

#define LOCKFILENAME "/tmp/connection_bar_lock"

int get_lock(int* fdlock)
{
    *fdlock = open( LOCKFILENAME, O_WRONLY | O_CREAT | O_CLOEXEC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
    if(*fdlock == -1)
    {
        syslog(LOG_ERR, "Failed to create lock file for connection bar %i", errno);
        return 0;
    }

    if(flock(*fdlock, LOCK_EX | LOCK_NB) == -1)
    {
        syslog(LOG_ERR, "Failed to lock file for connection bar %i", errno);
        return 0;
    }

    return 1;
}

void remove_lock(int* fdlock)
{
    close(*fdlock);
    remove(LOCKFILENAME);
}

void qrdesktop_close_handler(int)
{
    syslog(LOG_INFO, "Recieved SIGUSR1 signal from qrdesktop process.");
    QCoreApplication::exit(0);
}

int main(int argc, char* argv[])
{
    QApplication a(argc, argv);

    int fdlock;

    openlog("connectionbar", LOG_PID|LOG_NDELAY, LOG_USER);

    if (argc < 3)
    {
        syslog(LOG_ERR, "Connection bar started without session name, qrdesktopPID parameters");
        QCoreApplication::exit(1);
    }

    pid_t qrdesktopPID = atoi(argv[2]);
    if (qrdesktopPID == 0)
    {
        syslog(LOG_ERR, "Failed to convert qrdesktopPID parameter to int.");
        QCoreApplication::exit(1);
    }

    if (get_lock(&fdlock) != 0)
    {
        syslog(LOG_INFO, "Connection bar started with session name %s, qrdesktop PID %i", argv[1], qrdesktopPID);

        // SIGUSR1 = qrdesktop window is closed
        if (signal(SIGUSR1, qrdesktop_close_handler) == SIG_ERR)
        {
            syslog(LOG_WARNING, "Failed to set handler on SIGUSR1 signal.");
        }

        ConnectionBarWindow connectionBar(argv[1], qrdesktopPID);

        connectionBar.setWindowFlags(Qt::CustomizeWindowHint | Qt::WindowStaysOnTopHint | Qt::ToolTip);

        QRect screenRect = QApplication::desktop()->screenGeometry(-1);
        connectionBar.move((screenRect.width() / 2) - (connectionBar.width() / 2), -connectionBar.getMainLabelHeight());

        connectionBar.show();

        QrdesktopObserver observer(qrdesktopPID);
        observer.start();

        int application_result = a.exec();

        remove_lock(&fdlock);

        return application_result;
    }
    else
    {
        syslog(LOG_ERR, "Another instance of application is already running");
        QCoreApplication::exit(1);
    }
}
