#include "dynamicChannel.h"
#include "audioInputChannel.h"

#include <stdexcept>

// Uncomment this for debug hex output of messages
//#define DYNAMIC_CHANNEL_DEBUG

using namespace std;


DynamicChannelManager::DynamicChannelManager()
    :m_pTransportChannel(NULL)
{
}

void DynamicChannelManager::setTransportChannel(VCHANNEL* pTransportChannel)
{
    if (pTransportChannel != NULL)
    {
        m_pTransportChannel = pTransportChannel;
    }
    else
    {
        throw runtime_error("Dynamic Channel Manager DIDN'T start, transport channel is not created");
    }
}

void DynamicChannelManager::addDynamicChannel(DynamicChannel* pDynamicChannel)
{
    pDynamicChannel->setDynamicChannelManager(this);
    m_dynamicChannels.push_back(pDynamicChannel);
}

int DynamicChannelManager::readChannelId(STREAM s, uint8 header, uint32* p_channelId)
{
    uint8 cbChId = header & 3; // channel ID length flag
    int length = 0;

    switch (cbChId)
    {
        case 0: // channelId length = 1
        {
            in_uint8(s, *p_channelId);
            length = 1;
            break;
        }
        case 1: // channelId length = 2
        {
            in_uint16_le(s, *p_channelId);
            length = 2;
            break;
        }
        case 2: // channelId length = 4
        {
            in_uint32_le(s, *p_channelId);
            length = 4;
            break;
        }
        default: // invalid value
        {
            syslog(LOG_NOTICE, "Dynamic Channel Manager: wrong channel id length flag");
            break;;
        }
    }

    return length;
}

void DynamicChannelManager::sendCapabilitiesResponsePdu()
{
    uint8 cmd = CMD_CAPS << 4;
    uint16 capsVersion = 1; // For now we always send capabilities pdu with version = 1
    // 1b header + 1b pad + 2b version
    int messageLength = sizeof(cmd) + 1 + sizeof(capsVersion);

    STREAM outStream = channel_init(m_pTransportChannel, messageLength);
    out_uint8(outStream, cmd);
    out_uint8(outStream, 0); // pad 1 byte
    out_uint16_le(outStream, capsVersion);
    s_mark_end(outStream);
#ifdef DYNAMIC_CHANNEL_DEBUG
    hexdump(outStream->end - messageLength, messageLength);
#endif
    channel_send(outStream, m_pTransportChannel);
}

void DynamicChannelManager::sendCreateResponsePdu(uint32 channelId, uint32 creationStatus)
{
    uint8 cmd = CMD_CREATE << 4;
    uint8 cbChId = 2; // channelId length = 4
    uint8 header = cmd | cbChId;
    // 1 b header + 4 b ChannelId + 4 b CreationStatus
    int messageLength = sizeof(header) + sizeof(channelId) + sizeof(creationStatus);

    STREAM outStream = channel_init(m_pTransportChannel, messageLength);
    out_uint8(outStream, header);
    out_uint32_le(outStream, channelId);
    out_uint32_le(outStream, creationStatus);
    s_mark_end(outStream);
#ifdef DYNAMIC_CHANNEL_DEBUG
    hexdump(outStream->end - messageLength, messageLength);
#endif
    channel_send(outStream, m_pTransportChannel);
}

STREAM DynamicChannelManager::prepareDataPdu(uint32 channelId, int dataLength)
{
    uint8 cmd = CMD_DATA << 4;
    uint8 cbChId = 2; // channelId length = 4
    uint8 header = cmd | cbChId;
    // 1 b header + 4 b ChannelId + data length
    int messageLength = sizeof(header) + sizeof(channelId)  + dataLength;

    STREAM outStream = channel_init(m_pTransportChannel, messageLength);
    out_uint8(outStream, header);
    out_uint32_le(outStream, channelId);
    s_mark_end(outStream);

    return outStream;
}

void DynamicChannelManager::sendDataPdu(STREAM outStream)
{
#ifdef DYNAMIC_CHANNEL_DEBUG
    unsigned char* messageBegin = outStream->end; // saving previous end, end of dynamic channel header part of message
#endif
    s_mark_end(outStream);
#ifdef DYNAMIC_CHANNEL_DEBUG
    hexdump(messageBegin, outStream->end - messageBegin);
#endif
    channel_send(outStream, m_pTransportChannel);
}

void DynamicChannelManager::processCapabilitiesRequestPdu(STREAM s)
{
    sendCapabilitiesResponsePdu();
}

void DynamicChannelManager::processCreateRequestPdu(STREAM s)
{
    const uint32 STATUS_SUCCESS = 0;
    const uint32 STATUS_UNSUCCESSFUL = 0xc0000001;

    uint8 header; // message header
    in_uint8(s, header);

    uint32 channelId = 0;
    char length = readChannelId(s, header, &channelId);
    if (0 == length)
    {
        return;
    }

    char* pChannelName = reinterpret_cast<char*>(s->p);
    string channelName(pChannelName);
    for (int i = 0; i < m_dynamicChannels.size(); ++i)
    {
        if (m_dynamicChannels[i]->getName() == channelName)
        {
            m_dynamicChannels[i]->open(channelId);
            sendCreateResponsePdu(channelId, STATUS_SUCCESS);
            syslog(LOG_NOTICE, "Dynamic channel %s created", pChannelName);
            return;
        }        
    }
    sendCreateResponsePdu(channelId, STATUS_UNSUCCESSFUL);
}

void DynamicChannelManager::processCloseRequestPdu(STREAM s)
{
    uint8 header; // message header
    in_uint8(s, header);

    uint32 channelId = 0;
    char length = readChannelId(s, header, &channelId);
    if (0 == length)
    {
        return;
    }

    for (int i = 0; i < m_dynamicChannels.size(); ++i)
    {
        if (m_dynamicChannels[i]->isOpened() && m_dynamicChannels[i]->getId() == channelId)
        {
            m_dynamicChannels[i]->close();
            syslog(LOG_NOTICE, "Dynamic channel %s closed", m_dynamicChannels[i]->getName().c_str());
            break;
        }
    }
}

void DynamicChannelManager::processFirstDataPdu(STREAM s)
{
    uint8 header; // message header
    in_uint8(s, header);

    uint32 channelId = 0;
    char channelIdLength = readChannelId(s, header, &channelId);
    if (0 == channelIdLength)
    {
        return;
    }

    for (int i = 0; i < m_dynamicChannels.size(); ++i)
    {
        if (m_dynamicChannels[i]->isOpened() && m_dynamicChannels[i]->getId() == channelId)
        {
            unsigned char len = header & 12; // length of tatal message length (flag)
            uint32 totalLength;
            switch (len)
            {
                case 0: // length of total message length = 1
                {
                    in_uint8(s, totalLength);
                    break;
                }
                case 1: // length of total message length = 2
                {
                    in_uint16_le(s, totalLength);
                    break;
                }
                case 2: // length of total message length = 4
                {
                    in_uint32_le(s, totalLength);
                    break;
                }
                default: // invalid value
                {
                    return;
                }
            }
            break;
        }
    }
    // TO DO !!!
    // implement data reading to somewhere, buffer or something
    // currentDataLength = MAX_DYNAMIC_CHANNEL_DATA_LENGTH - 1 - channelIdLength - totalLengthSize;
}

void DynamicChannelManager::processDataPdu(STREAM s)
{
    uint8 header; // message header
    in_uint8(s, header);

    uint32 channelId = 0;
    char length = readChannelId(s, header, &channelId);
    if (0 == length)
    {
        return;
    }

    for (int i = 0; i < m_dynamicChannels.size(); ++i)
    {
        if (m_dynamicChannels[i]->isOpened() && m_dynamicChannels[i]->getId() == channelId)
        {
            m_dynamicChannels[i]->processMessage(s, 0);
            break;
        }
    }
    // TO DO !!!
    // implement if this is the part of long message
}


//----------------
// DynamicChannel

DynamicChannel::DynamicChannel( const std::string& name)
    : m_pDynamicChannelManager(NULL), m_name(name), m_isOpened(false)
{
}

void DynamicChannel::setDynamicChannelManager(DynamicChannelManager* pDynamicChannelManager)
{
    if (pDynamicChannelManager != NULL)
    {
        m_pDynamicChannelManager = pDynamicChannelManager;
    }
    else
    {
        string errorString(" Channel WILL NOT start, Dynamic Channel Manager is not created");
        errorString = m_name.c_str() + errorString;
        throw runtime_error(errorString);
    }
}

std::string& DynamicChannel::getName()
{
    return m_name;
}

uint32 DynamicChannel::getId()
{
    return m_id;
}

bool DynamicChannel::isOpened()
{
    return m_isOpened;
}

void DynamicChannel::open(uint32 id)
{
    m_id = id;
    m_isOpened = true;
}

void DynamicChannel::close()
{
    onClose();
    m_id = 0;
    m_isOpened = false;
}


// --------------
// Init functions

static VCHANNEL* g_pTransportChannel = NULL;
static DynamicChannelManager g_dynamicChannelManager;
static AudioInputChannel g_audioInputChannel;


void dynamicChannelProcess(STREAM s, char* name)
{
#ifdef DYNAMIC_CHANNEL_DEBUG
    hexdump(s->p, s->end - s->p);
#endif
    uint8 cmd = s->p[0] >> 4; // Cmd, command from header

    switch (cmd)
    {
        case DynamicChannelManager::CMD_CAPS:
        {
            g_dynamicChannelManager.processCapabilitiesRequestPdu(s);
            break;
        }
        case DynamicChannelManager::CMD_CREATE:
        {
            g_dynamicChannelManager.processCreateRequestPdu(s);
            break;
        }
        case DynamicChannelManager::CMD_DATA:
        {
            g_dynamicChannelManager.processDataPdu(s);
            syslog(LOG_NOTICE, "=== process channel DRDYNVC, data");
            break;
        }
        case DynamicChannelManager::CMD_DATA_FIRST:
        {
            g_dynamicChannelManager.processFirstDataPdu(s);
            syslog(LOG_NOTICE, "=== process channel DRDYNVC, data first");
            break;
        }
        case DynamicChannelManager::CMD_CLOSE:
        {
            g_dynamicChannelManager.processCloseRequestPdu(s);
            break;
        }
        default:
        {
            break;
        }
    }
}

void dynamic_channel_init()
{
    g_pTransportChannel =
          channel_register("DRDYNVC", CHANNEL_OPTION_INITIALIZED | CHANNEL_OPTION_ENCRYPT_RDP, dynamicChannelProcess);

    try
    {
        g_dynamicChannelManager.setTransportChannel(g_pTransportChannel);
        g_dynamicChannelManager.addDynamicChannel(&g_audioInputChannel);
        syslog(LOG_NOTICE, "Dynamic channel registered");
    }
    catch(runtime_error& error)
    {
        syslog(LOG_ERR, error.what());
    }
}
