#include "pnsslcli.h"
#include <syslog.h>

#include <boost/bind.hpp>
#include <assert.h>
#include <vector>

// ------------------------------------------------------------------------------
// session member functions realization
// ------------------------------------------------------------------------------
session::session (boost::asio::io_service& io_service,
                  boost::asio::ssl::context& context)
    : socket_ (io_service),
      ssl_socket_ (io_service, context),
      timer_ (io_service)
{
}

// ------------------------------------------------------------------------------
tcp::socket& session::socket ()
{
    return socket_;
}

// ------------------------------------------------------------------------------
ssl_socket& session::sslsocket ()
{
    return ssl_socket_;
}

// ------------------------------------------------------------------------------
void session::start (const std::string& rdphost, const std::string& username, const std::string& ticket)
{
    std::stringstream connect_message;

    const char* connect_str = "CONNECT:";
    const char* client_str  = "CLIENT:";
    const char* ticket_str  = "TICKET:";
    const char* crlf_str    = "\r\n";
    const char* space_str   = " ";

    // Compose message wih format "CONNECT: %s\r\nCLIENT: %s\r\nTICKET: %s\r\n\r\n"
    connect_message << connect_str;
    connect_message << space_str;
    connect_message << rdphost;
    connect_message << crlf_str;
    connect_message << client_str;
    connect_message << space_str;
    connect_message << username;
    connect_message << crlf_str;
    if (!ticket.empty())
    {
        // Force Broker Authentication
        // add ticket to message "CONNECT: %s\r\nCLIENT: %s\r\nTICKET: %s\r\n\r\n"
        connect_message << ticket_str;
        connect_message << space_str;
        connect_message << ticket;
        connect_message << crlf_str;
    }
    connect_message << crlf_str;

    syslog (LOG_NOTICE, "%s", connect_message.str().c_str());
    ssl_socket_.write_some (boost::asio::buffer (connect_message.str().c_str(), connect_message.str().size()));

    ssl_socket_.async_read_some (boost::asio::buffer (data_in_),
        make_custom_alloc_handler (allocator_in_,
            boost::bind (&session::handle_server_read,
               shared_from_this (),
               boost::asio::placeholders::error,
               boost::asio::placeholders::bytes_transferred)));
    socket_.async_read_some (boost::asio::buffer (data_out_),
        make_custom_alloc_handler (allocator_out_,
            boost::bind (&session::handle_read,
                   shared_from_this (),
                   boost::asio::placeholders::error,
                   boost::asio::placeholders::bytes_transferred)));
}

// ------------------------------------------------------------------------------
void session::on_timer (const boost::system::error_code& e)
{
    if (!e)
    {
        syslog (LOG_NOTICE, "write operation timeout");
        ssl_socket_.lowest_layer().close();
        socket_.lowest_layer().shutdown(boost::asio::ip::tcp::socket::shutdown_both);
    }
}

// ------------------------------------------------------------------------------
void session::handle_read (const boost::system::error_code & error, size_t bytes_transferred)
{
    if (!error)
    {
        timer_.expires_from_now (boost::posix_time::seconds(10));
        timer_.async_wait (boost::bind (&session::on_timer,
                                        shared_from_this (),
                                        boost::asio::placeholders::error));
        boost::asio::async_write (ssl_socket_,
                boost::asio::buffer (data_out_, bytes_transferred),
                make_custom_alloc_handler (allocator_out_,
                    boost::bind (&session::handle_server_write,
                        shared_from_this (),
                        boost::asio::placeholders::error)));
    }
    else
    {
      throw error;
    }
}

// ------------------------------------------------------------------------------
void session::handle_server_read (const boost::system::error_code & error, size_t bytes_transferred)
{
    if (!error)
    {
        boost::asio::async_write (socket_,
                boost::asio::buffer (data_in_, bytes_transferred),
                make_custom_alloc_handler (allocator_in_,
                    boost::bind (&session::handle_write,
                        shared_from_this (),
                        boost::asio::placeholders::error)));
    }
    else
    {
      syslog (LOG_NOTICE, "server_read error: %s, bytes read: %d", error.message().c_str(), bytes_transferred);
      throw error;
    }
}

// ------------------------------------------------------------------------------
void session::handle_write (const boost::system::error_code & error)
{
    if (!error)
    {
        ssl_socket_.async_read_some (boost::asio::buffer (data_in_),
            make_custom_alloc_handler (allocator_in_,
                boost::bind (&session::handle_server_read,
                    shared_from_this (),
                    boost::asio::placeholders::error,
                    boost::asio::placeholders::bytes_transferred)));
    }
    else
    {
      throw error;
    }
}

// ------------------------------------------------------------------------------
void session::handle_server_write (const boost::system::error_code & error)
{
    timer_.cancel();
    if (!error)
    {
        socket_.async_read_some (boost::asio::buffer (data_out_),
                make_custom_alloc_handler (allocator_out_,
                    boost::bind (&session::handle_read,
                        shared_from_this (),
                        boost::asio::placeholders::error,
                        boost::asio::placeholders::bytes_transferred)));
    }
    else
    {
      throw error;
    }
}


// ------------------------------------------------------------------------------
//                      pnssl_server
// ------------------------------------------------------------------------------

pnssl_server::pnssl_server(boost::asio::io_service& io_service,
                           const tcp::endpoint& endpoint,
                           tcp::resolver::iterator & iterator,
                           const char * remote_addr,
                           const char * username,
                           const char * ticket):
                                                io_service_(io_service),
                                                acceptor_(io_service, endpoint),
                                                endpoint_iterator_(iterator),
                                                remote_addr_(remote_addr),
                                                username_(username),
                                                ticket_(ticket)
{
    boost::asio::ssl::context ctx(io_service_, boost::asio::ssl::context::sslv23);
    session_ptr new_session(new session(io_service_, ctx));
    ctx.set_verify_mode (boost::asio::ssl::context::verify_peer);
    acceptor_.async_accept(new_session->socket(),
                           boost::bind(&pnssl_server::handle_accept, this, new_session,
                                       boost::asio::placeholders::error));
}

void pnssl_server::handle_accept(session_ptr sess,
                   const boost::system::error_code& error)
{
    if (!error)
    {
        // connect to SSL VPN server
        sess->sslsocket().lowest_layer().connect(*endpoint_iterator_);
        syslog (LOG_NOTICE, "pnssl_server: connect to ssl gateway is done");
        sess->sslsocket().handshake(boost::asio::ssl::stream_base::client);
        syslog (LOG_NOTICE, "pnssl_server: handshake passed");
        sess->start(remote_addr_.c_str(), username_.c_str(), ticket_.c_str());

        boost::asio::ssl::context ctx(io_service_, boost::asio::ssl::context::sslv23);
        session_ptr new_session(new session(io_service_, ctx));
        ctx.set_verify_mode (boost::asio::ssl::context::verify_peer);
        acceptor_.async_accept(new_session->socket(),
                               boost::bind(&pnssl_server::handle_accept, this, new_session,
                                           boost::asio::placeholders::error));
    }
}

