#include "pnsslcli.h"
#include <cstdlib>
#include <syslog.h>
#include <sys/wait.h>
#include <signal.h>
#include <cassert>
#include <vector>
#include <pthread.h>

#include "proxy_utils.h"

extern char g_localaddr[64];
extern char g_ssladdr[64];
extern char g_ssladdr_local[64];
extern char g_remoteaddr[64];
extern char g_username[64];
extern const char g_secure_ticket_filename[_POSIX_NAME_MAX];
extern int g_send_secure_ticket;
std::vector<char> g_ticket_buf;

// ------------------------------------------------------------------------------
//  main() arguments:
//  1) local loopback address <host[:port]>, default port is 3389
//  2) SSL VPN server address <host[:port]>, default port is 443
//  3) RDP hostname and 4) username, they're passed to VPN server
//      5) PNTSC pid, to say to that initialization is finished
// ------------------------------------------------------------------------------

class FileCloserGuard
{
public:

    FileCloserGuard(FILE* pFile)
    :m_pFile(pFile)
    {
    }

    ~FileCloserGuard()
    {
        pclose(m_pFile);
    }

private:

    FileCloserGuard(FileCloserGuard& other);
    FileCloserGuard& operator = (FileCloserGuard);

    FILE* m_pFile;
};

static void read_ticket(std::vector<char>* ticket_buf)
{
    FILE* pFile = fopen(g_secure_ticket_filename, "rb");
    if (pFile != NULL)
    {
        FileCloserGuard fileGuard(pFile);

        // obtain file size:
        fseek(pFile , 0 , SEEK_END);
        long lSize = ftell(pFile);
        rewind(pFile);

        ticket_buf->resize((size_t)lSize);
        size_t result = ::fread(&(*ticket_buf)[0], 1, ticket_buf->size(), pFile);
        if (result != ticket_buf->size())
        {
            syslog(LOG_WARNING, "Cannot read secure ticket");
        }
        // Do not delete file here, pnkerb handler will read ticket
    }
}
void unlock_ssl_mutex(pthread_mutex_t * ssl_mutex)
{
    if (pthread_mutex_trylock(ssl_mutex) == EBUSY)
    {
        pthread_mutex_unlock(ssl_mutex);
    }
}

extern "C"
void *
pnsslcli_thread(pthread_mutex_t * ssl_mutex)
{
    syslog (LOG_NOTICE, "SSL start session");

    try
    {
        boost::asio::io_service io_service;

        // connect to SSL VPN server
        std::string ssladdr(g_ssladdr);
        std::string sslport("443"); // default
        size_t portPosition = 0;
        if (std::string::npos != (portPosition = ssladdr.find(":")))
        {
            sslport.assign(ssladdr.begin() + portPosition + 1, ssladdr.end());
            ssladdr.resize(portPosition);
        }

        tcp::resolver resolver (io_service);
        tcp::resolver::query query (tcp::v4 (), ssladdr.c_str(), sslport.c_str());
        tcp::resolver::iterator iterator = resolver.resolve (query);
        
        syslog (LOG_NOTICE, "SSL gateway: %s:%s", ssladdr.c_str(), sslport.c_str());

        // listen to local loopback
        std::string localaddr(g_localaddr);
        unsigned short startPort = 3389; // default

        if (std::string::npos != (portPosition = localaddr.find(":")))
        {
            std::string port(localaddr.begin() + portPosition + 1, localaddr.end());
            startPort = std::atoi(port.c_str());
            if (startPort == 0)
            {
                startPort = 3389;
            }
            localaddr.resize(portPosition);
        }

        const int portsDelta = 100;
        const unsigned short endPort = (startPort < (USHRT_MAX - portsDelta)) ? startPort + portsDelta : USHRT_MAX;
        
        unsigned short port = getFreePort(startPort, endPort);
        
        tcp::endpoint endpoint(boost::asio::ip::address::from_string (localaddr), port);

        if (g_send_secure_ticket)
        {
            read_ticket(&g_ticket_buf);
        }
        else
        {
            g_ticket_buf.clear();
        }
        std::string secure_ticket(g_ticket_buf.begin(), g_ticket_buf.end());

        server_ptr server(new pnssl_server(io_service, endpoint, iterator, g_remoteaddr, g_username, secure_ticket.c_str()));

        syslog (LOG_NOTICE, "remote server: %s", g_remoteaddr);
        snprintf(g_ssladdr_local, sizeof(g_ssladdr_local), "%s:%d", localaddr.c_str(), port);
        syslog (LOG_NOTICE, "SSL proxy: %s", g_ssladdr_local);

        unlock_ssl_mutex(ssl_mutex);

        io_service.run ();


    }
    catch (const boost::system::error_code & e)
    {
        unlock_ssl_mutex(ssl_mutex);
        syslog (LOG_NOTICE, "pnsslcli_thread: error '%s'", e.message().c_str());
        // SSL Gateway rejected secure ticket. We'll try to connect without it.
        return 0;
    }
    catch (std::exception & e)
    {
        unlock_ssl_mutex(ssl_mutex);
        syslog (LOG_NOTICE, "pnsslcli_thread: an exception was thrown: %s\n", e.what ());
    }
    catch (...)
    {
        unlock_ssl_mutex(ssl_mutex);
        syslog (LOG_NOTICE, "pnsslcli_thread: an unknown exception was thrown\n");
    }

    syslog (LOG_NOTICE, "pnsslcli_thread: session was closed");
    closelog ();
    pthread_exit(NULL);
}
